# pyOCD debugger
# Copyright (c) 2020 Cypress Semiconductor Corporation
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...core.memory_map import (RamRegion, RomRegion)
from ...flash.flash import Flash

LOG = logging.getLogger(__name__)


class Flash_PSoC64(Flash):
    isFlashing = False

    def init(self, operation, address=None, clock=0, reset=True):
        if self._active_operation != operation and self._active_operation is not None:
            self.uninit()

        Flash_PSoC64.isFlashing = True
        super(Flash_PSoC64, self).init(operation, address, clock, reset)
        Flash_PSoC64.isFlashing = True
        LOG.debug("Flash_PSoC64: initialised for %s", operation)

    def uninit(self):
        if self._active_operation is None:
            return

        super(Flash_PSoC64, self).uninit()
        Flash_PSoC64.isFlashing = False
        LOG.debug("Flash_PSoC64: uninitialised")


class PSoC6FlashParams:
    # Main/Work Flash flash operation weights
    MFLASH_ERASE_ALL_WEIGHT = 0.5
    MFLASH_ERASE_SECTOR_WEIGHT = 0.05
    MFLASH_PROGRAM_PAGE_WEIGHT = 0.07

    # External (SMIF) Flash flash operation weights
    SMIF_ERASE_ALL_WEIGHT = 140
    SMIF_ERASE_SECTOR_WEIGHT = 1
    SMIF_PROGRAM_PAGE_WEIGHT = 0.5

    defaultRomRegion = RomRegion(start=0x00000000, length=0x20000)
    defaultRamRegion = RamRegion(start=0x08000000, length=0x8000)



# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .cortex_m import CortexM
from .core_ids import (CORE_TYPE_NAME, CoreArchitecture, CortexMExtension)
from ..core.target import Target
from .cortex_m_core_registers import CoreRegisterGroups

LOG = logging.getLogger(__name__)

class CortexM_v8M(CortexM):
    """@brief Component class for a v8.x-M architecture Cortex-M core."""

    ARMv8M_BASE = 0xC
    ARMv8M_MAIN = 0xF

    ## DFSR.PMU added in v8.1-M.
    DFSR_PMU = (1 << 5)

    DSCSR = 0xE000EE08
    DSCSR_CDSKEY = 0x00020000
    DSCSR_CDS = 0x00010000
    DSCSR_SBRSEL = 0x00000002
    DSCSR_SBRSELEN = 0x00000001

    # Processor Feature Register 0
    PFR0 = 0xE000ED40
    PFR0_RAS_MASK = 0xf0000000
    PFR0_RAS_SHIFT = 28
    PFR0_RAS_VERSION_1 = 2

    # Processor Feature Register 1
    PFR1 = 0xE000ED44
    PFR1_SECURITY_MASK = 0x000000f0
    PFR1_SECURITY_SHIFT = 4

    PFR1_SECURITY_EXT_V8_0 = 0x1 # Base security extension.
    PFR1_SECURITY_EXT_V8_1 = 0x3 # v8.1-M adds several instructions.

    # Debug Feature Register 0
    DFR0 = 0xE000ED48
    DFR0_UDE_MASK = 0xf0000000
    DFR0_UDE_SHIFT = 28
    DFR0_UDE_SUPPORTED = 1

    # Media and FP Feature Register 1
    MVFR1 = 0xE000EF44
    MVFR1_MVE_MASK = 0x00000f00
    MVFR1_MVE_SHIFT = 8
    MVFR1_MVE__INTEGER = 0x1
    MVFR1_MVE__FLOAT = 0x2
    MVFR1_FP16_MASK = 0x00f00000
    MVFR1_FP16_SHIFT = 20
    MVFR1_FP16__SUPPORTED = 0x1 # FP16 format support is present.

    # Instruction Set Attribute Register 0
    ISAR0 = 0xE000ED60
    ISAR0_CMPBRANCH_MASK = 0x0000f000
    ISAR0_CMPBRANCH_SHIFT = 12
    ISAR0_CMPBRANCH__LOB = 0x3 # LOB instructions from v8.1-M are present.

    # Instruction Set Attribute Register 5
    ISAR5 = 0xE000ED74
    ISAR5_PACBTI_MASK = 0x00f00000
    ISAR5_PACBTI_SHIFT = 20
    ISAR5_PACBTI__NONE = 0x0 # PACBTI is not present.

    # PMU Type register
    PMU_TYPE = 0xE0003E00
    PMU_TYPE_N_MASK  = 0x0000000f

    def __init__(self, rootTarget, ap, memory_map=None, core_num=0, cmpid=None, address=None):
        super().__init__(rootTarget, ap, memory_map, core_num, cmpid, address)

    @property
    def supported_security_states(self):
        """@brief Tuple of security states supported by the processor.

        @return Tuple of @ref pyocd.core.target.Target.SecurityState "Target.SecurityState". The
            result depends on whether the Security extension is enabled.
        """
        if self.has_security_extension:
            return (Target.SecurityState.NONSECURE, Target.SecurityState.SECURE)
        else:
            return (Target.SecurityState.NONSECURE,)

    def _read_core_type(self):
        """@brief Read the CPUID register and determine core type and architecture."""
        # Schedule deferred reads.
        cpuid_cb = self.read32(self.CPUID, now=False)
        pfr0_cb = self.read32(self.PFR0, now=False)
        pfr1_cb = self.read32(self.PFR1, now=False)
        dfr0_cb = self.read32(self.DFR0, now=False)
        isar0_cb = self.read32(self.ISAR0, now=False)
        isar3_cb = self.read32(self.ISAR3, now=False)
        isar5_cb = self.read32(self.ISAR5, now=False)
        pmu_type_cb = self.read32(self.PMU_TYPE, now=False)
        mpu_type_cb = self.read32(self.MPU_TYPE, now=False)

        # Read CPUID register
        cpuid = cpuid_cb()
        implementer = (cpuid & CortexM.CPUID_IMPLEMENTER_MASK) >> CortexM.CPUID_IMPLEMENTER_POS
        arch = (cpuid & CortexM.CPUID_ARCHITECTURE_MASK) >> CortexM.CPUID_ARCHITECTURE_POS
        self.core_type = (cpuid & CortexM.CPUID_PARTNO_MASK) >> CortexM.CPUID_PARTNO_POS
        self.cpu_revision = (cpuid & CortexM.CPUID_VARIANT_MASK) >> CortexM.CPUID_VARIANT_POS
        self.cpu_patch = (cpuid & CortexM.CPUID_REVISION_MASK) >> CortexM.CPUID_REVISION_POS

        # Check for DSP extension
        isar3 = isar3_cb()
        isar3_simd = (isar3 & self.ISAR3_SIMD_MASK) >> self.ISAR3_SIMD_SHIFT
        if isar3_simd == self.ISAR3_SIMD__DSP:
            self._extensions.append(CortexMExtension.DSP)

        # Check for RAS extension.
        pfr0 = pfr0_cb()
        pfr0_ras = (pfr0 & self.PFR0_RAS_MASK) >> self.PFR0_RAS_SHIFT
        if pfr0_ras == self.PFR0_RAS_VERSION_1:
            self._extensions.append(CortexMExtension.RAS)

        # Check for the security extension.
        pfr1 = pfr1_cb()
        pfr1_sec = (pfr1 & self.PFR1_SECURITY_MASK) >> self.PFR1_SECURITY_SHIFT
        self.has_security_extension = pfr1_sec in (self.PFR1_SECURITY_EXT_V8_0, self.PFR1_SECURITY_EXT_V8_1)
        if self.has_security_extension:
            self._extensions.append(CortexMExtension.SEC)
        if pfr1_sec == self.PFR1_SECURITY_EXT_V8_1:
            self._extensions.append(CortexMExtension.SEC_V81)

        # Check for UDE extension.
        dfr0 = dfr0_cb()
        dfr0_ude = (dfr0 & self.DFR0_UDE_MASK) >> self.DFR0_UDE_SHIFT
        if dfr0_ude == self.DFR0_UDE_SUPPORTED:
            self._extensions.append(CortexMExtension.UDE)

        # Check for PACBTI extension.
        isar5 = isar5_cb()
        isar5_pacbti = (isar5 & self.ISAR5_PACBTI_MASK) >> self.ISAR5_PACBTI_SHIFT
        if isar5_pacbti != self.ISAR5_PACBTI__NONE:
            self._extensions.append(CortexMExtension.PACBTI)

        # Check for PMU extension.
        pmu_type = pmu_type_cb()
        pmu_type_n = pmu_type & self.PMU_TYPE_N_MASK
        if pmu_type_n > 0:
            self._extensions.append(CortexMExtension.PMU)

        # Check for MPU extension
        mpu_type = mpu_type_cb()
        mpu_type_dregions = (mpu_type & self.MPU_TYPE_DREGIONS_MASK) >> self.MPU_TYPE_DREGIONS_SHIFT
        if mpu_type_dregions > 0:
            self._extensions.append(CortexMExtension.MPU)

        # Determine the base/main variant.
        if arch == self.ARMv8M_BASE:
            self._architecture = CoreArchitecture.ARMv8M_BASE
        else:
            self._architecture = CoreArchitecture.ARMv8M_MAIN

        # Determine the architecture major/minor version.
        # The presence of low-overhead loop and branch instructions is used to distinguish v8.1-M from v8.0-M.
        isar0 = isar0_cb()
        isar0_cmpbranch = (isar0 & self.ISAR0_CMPBRANCH_MASK) >> self.ISAR0_CMPBRANCH_SHIFT
        if isar0_cmpbranch == self.ISAR0_CMPBRANCH__LOB:
            self._arch_version = (8, 1)
        else:
            self._arch_version = (8, 0)

        self._core_name = CORE_TYPE_NAME.get((implementer, self.core_type), f"Unknown (CPUID={cpuid:#010x})")

    def _check_for_fpu(self):
        """@brief Determine if a core has an FPU.

        In addition to the tests performed by CortexM, this method tests for the MVE extension.
        """
        # Schedule this deferred read before calling the super implementation.
        mvfr1_cb = self.read32(self.MVFR1, now=False)

        super()._check_for_fpu()

        # Check for MVE.
        mvfr1 = mvfr1_cb()
        mve = (mvfr1 & self.MVFR1_MVE_MASK) >> self.MVFR1_MVE_SHIFT
        if mve == self.MVFR1_MVE__INTEGER:
            self._extensions.append(CortexMExtension.MVE)
        elif mve == self.MVFR1_MVE__FLOAT:
            self._extensions += [CortexMExtension.MVE, CortexMExtension.MVE_FP]

        # Check for half-precision FP.
        fp16 = (mvfr1 & self.MVFR1_FP16_MASK) >> self.MVFR1_FP16_SHIFT
        if fp16 == self.MVFR1_FP16__SUPPORTED:
            self._extensions.append(CortexMExtension.FPU_HP)

    def _build_registers(self):
        super()._build_registers()

        # Registers available with Security extension, either Baseline or Mainline.
        if self.has_security_extension:
            self._core_registers.add_group(CoreRegisterGroups.V8M_SEC_ONLY)

        # Mainline-only registers.
        if self.architecture == CoreArchitecture.ARMv8M_MAIN:
            self._core_registers.add_group(CoreRegisterGroups.V7M_v8M_ML_ONLY)

            # Registers available when both Mainline and Security extensions are implemented.
            if self.has_security_extension:
                self._core_registers.add_group(CoreRegisterGroups.V8M_ML_SEC_ONLY)

        # MVE registers.
        if CortexMExtension.MVE in self.extensions:
            self._core_registers.add_group(CoreRegisterGroups.V81M_MVE_ONLY)

    def get_security_state(self):
        """@brief Returns the current security state of the processor.

        @return @ref pyocd.core.target.Target.SecurityState "Target.SecurityState" enumerator.
        """
        dscsr = self.read32(self.DSCSR)
        if (dscsr & self.DSCSR_CDS) != 0:
            return Target.SecurityState.SECURE
        else:
            return Target.SecurityState.NONSECURE

    def clear_debug_cause_bits(self):
        self.write32(CortexM.DFSR,
                self.DFSR_PMU
                | CortexM.DFSR_EXTERNAL
                | CortexM.DFSR_VCATCH
                | CortexM.DFSR_DWTTRAP
                | CortexM.DFSR_BKPT
                | CortexM.DFSR_HALTED)

    def get_halt_reason(self):
        """@brief Returns the reason the core has halted.

        This overridden version of this method adds support for v8.x-M halt reasons.

        @return @ref pyocd.core.target.Target.HaltReason "Target.HaltReason" enumerator or None.
        """
        dfsr = self.read32(self.DFSR)
        if dfsr & self.DFSR_HALTED:
            reason = Target.HaltReason.DEBUG
        elif dfsr & self.DFSR_BKPT:
            reason = Target.HaltReason.BREAKPOINT
        elif dfsr & self.DFSR_DWTTRAP:
            reason = Target.HaltReason.WATCHPOINT
        elif dfsr & self.DFSR_VCATCH:
            reason = Target.HaltReason.VECTOR_CATCH
        elif dfsr & self.DFSR_EXTERNAL:
            reason = Target.HaltReason.EXTERNAL
        elif dfsr & self.DFSR_PMU:
            reason = Target.HaltReason.PMU
        else:
            reason = None
        return reason




# pyOCD debugger
# Copyright (c) 2018,2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .component import CoreSightComponent
from ..utility.timeout import Timeout

ACK_TIMEOUT = 5.0

class GPR(CoreSightComponent):
    """@brief Granular Power Requestor.

    Currently only supports enabling power domains.
    """

    CPWRUPREQ = 0x0
    CPWRUPACK = 0x0

    CPWRUPM_COUNT_MASK = 0x3f

    @classmethod
    def factory(cls, ap, cmpid, address):
        # Attempt to return the same instance that was created during ROM table scanning.
        if cmpid.parent_rom_table is not None:
            rom_gpr = cmpid.parent_rom_table.gpr
            if rom_gpr is not None and rom_gpr.address == address:
                return rom_gpr

        # No luck, create a new instance.
        gpr = cls(ap, cmpid, address)
        return gpr

    def __init__(self, ap, cmpid=None, addr=None):
        super(GPR, self).__init__(ap, cmpid, addr)
        self.domain_count = 0

    def init(self):
        """@brief Inits the GPR."""
        self.domain_count = self.cmpid.devid[2] & self.CPWRUPM_COUNT_MASK

    def _power_up(self, mask):
        """@brief Enable power to a power domaind by mask.
        @param self
        @param mask Bitmask of the domains to power up.
        @retval True Requested domains were successfully powered on.
        @return False Timeout waiting for power ack bit(s) to set.
        """
        # Enable power up request bits.
        self.ap.write32(self.address + self.CPWRUPREQ, mask)

        # Wait for ack bits to set.
        with Timeout(ACK_TIMEOUT) as t_o:
            while t_o.check():
                value = self.ap.read32(self.address + self.CPWRUPACK)
                if (value & mask) == mask:
                    return True
            return False

    def power_up_all(self):
        """@brief Enable power to all available power domains.
        @param self
        @retval True All domains were successfully powered on.
        @return False Timeout waiting for power ack bit(s) to set.
        """
        mask = (1 << self.domain_count) - 1
        return self._power_up(mask)

    def power_up_one(self, domain_id):
        """@brief Power up a single power domain by domain ID.
        @param self
        @param domain_id Integer power domain ID.
        @retval True Requested domain was powered on successfully.
        @return False Timeout waiting for power ack bit to set.
        """
        mask = 1 << domain_id
        return self._power_up(mask)

    def __repr__(self):
        return "<GPR @ %x: count=%d>" % (id(self), self.domain_count)






# pyOCD debugger
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import (TYPE_CHECKING, Optional, Set)

from .scope import Scope

LOG = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .sequences import (DebugSequence, DebugSequenceExecutionContext)
    from ...target.pack.cmsis_pack import CmsisPackDevice

class DebugSequenceDelegate:
    """@brief Delegate interface for handling sequence operations."""

    @property
    def all_sequences(self) -> Set[DebugSequence]:
        """@brief Returns a set containing all defined debug sequence objects."""
        raise NotImplementedError()

    @property
    def cmsis_pack_device(self) -> CmsisPackDevice:
        """@brief Accessor for the pack device that contains the sequences."""
        raise NotImplementedError()

    def get_root_scope(self, context: DebugSequenceExecutionContext) -> Scope:
        """@brief Return a scope that will be used as the parent of sequences.

        Normally the delegate will return the debugvars scope from this method. It's also possible to
        simply return an empty scope.
        """
        raise NotImplementedError()

    def run_sequence(self, name: str, pname: Optional[str] = None) -> Optional[Scope]:
        """@brief Execute the debug sequence with the specified name.
        @exception NameError No sequence with the given name is defined.
        """
        raise NotImplementedError()

    def has_sequence_with_name(self, name: str, pname: Optional[str] = None) -> bool:
        """@brief Return whether there is a debug sequence with the specified name."""
        raise NotImplementedError()

    def get_sequence_with_name(self, name: str, pname: Optional[str] = None) -> DebugSequence:
        """@brief Return the named debug sequence object.

        Expected to raise if the sequence isn't available.
        """
        raise NotImplementedError()

    def get_protocol(self) -> int:
        """@brief Return the value for the __protocol variable.
        __protocol fields:
        - [15:0] 0=error, 1=JTAG, 2=SWD, 3=cJTAG
        - [16] SWJ-DP present?
        - [17] switch through dormant state?
        """
        raise NotImplementedError()

    def get_connection_type(self) -> int:
        """@brief Return the value for the __connection variable.
        __connection fields:
        - [7:0] connection type: 0=error/disconnected, 1=for debug, 2=for flashing
        - [15:8] reset type: 0=error, 1=hw, 2=SYSRESETREQ, 3=VECTRESET
        - [16] connect under reset?
        - [17] pre-connect reset?
        """
        raise NotImplementedError()

    def get_traceout(self) -> int:
        """@brief Return the value for the __traceout variable.
        __traceout fields:
        - [0] SWO enabled?
        - [1] parallel trace enabled?
        - [2] trace buffer enabled?
        - [21:16] selected parallel trace port size
        """
        raise NotImplementedError()

    def get_sequence_functions(self) -> DebugSequenceFunctionsDelegate:
        """@brief Return an instance of the sequence function implementations delegate.

        This method lets the delegate determine the set of built-in sequence functions.
        """
        raise NotImplementedError()

class DebugSequenceFunctionsDelegate:
    """@brief Implements functions provided by the debug sequence environment.

    All defined functions must have type annotations. Any function that has a fixed return value of 0
    should return None. This will be converted to 0 by the interpreter.

    The function names must be all lower-case in order to support case-insensitive symbol lookup.
    Whether this is actually correct debug sequence behaviour is unknown since it's not documented
    in the Open-CMSIS-Pack specification (version 1.7.15) as of late December 2022.
    """

    @property
    def context(self) -> DebugSequenceExecutionContext:
        from .sequences import DebugSequenceExecutionContext
        return DebugSequenceExecutionContext.get_active_context()



# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import (Dict, Tuple)

# pylint: disable=invalid_name

# CPUID IMPLEMENTER values
CPUID_ARM = 0x41
CPUID_ARM_CHINA = 0x63

# CPUID PARTNO values
ARM_SC000 = 0xC30
ARM_SC300 = 0xC33
ARM_CortexM0 = 0xC20
ARM_CortexM1 = 0xC21
ARM_CortexM3 = 0xC23
ARM_CortexM4 = 0xC24
ARM_CortexM7 = 0xC27
ARM_CortexM0p = 0xC60
ARM_CortexM23 = 0xD20
ARM_CortexM33 = 0xD21
ARM_CortexM35P = 0xD31
ARM_CortexM55 = 0xD22
ARM_CortexM85 = 0xD23
ARM_China_StarMC1 = 0x132

# pylint: enable=invalid_name

## @brief User-friendly names for core types.
CORE_TYPE_NAME: Dict[Tuple[int, int], str] = {
        (CPUID_ARM,        ARM_SC000):         "SecurCore SC000",
        (CPUID_ARM,        ARM_SC300):         "SecurCore SC300",
        (CPUID_ARM,        ARM_CortexM0):      "Cortex-M0",
        (CPUID_ARM,        ARM_CortexM1):      "Cortex-M1",
        (CPUID_ARM,        ARM_CortexM3):      "Cortex-M3",
        (CPUID_ARM,        ARM_CortexM4):      "Cortex-M4",
        (CPUID_ARM,        ARM_CortexM7):      "Cortex-M7",
        (CPUID_ARM,        ARM_CortexM0p):     "Cortex-M0+",
        (CPUID_ARM,        ARM_CortexM23):     "Cortex-M23",
        (CPUID_ARM,        ARM_CortexM33):     "Cortex-M33",
        (CPUID_ARM,        ARM_CortexM35P):    "Cortex-M35P",
        (CPUID_ARM,        ARM_CortexM55):     "Cortex-M55",
        (CPUID_ARM,        ARM_CortexM85):     "Cortex-M85",
        (CPUID_ARM_CHINA,  ARM_China_StarMC1): "Star-MC1",
    }

class CoreArchitecture(Enum):
    """@brief CPU architectures."""
    ARMv6M = 1
    ARMv7M = 2
    ARMv8M_BASE = 3
    ARMv8M_MAIN = 4

class CortexMExtension(Enum):
    """@brief Extensions for the Cortex-M architecture."""
    FPU = "FPU" # Single-Precision floating point
    DSP = "DSP" # Digital Signal Processing instructions
    FPU_DP = "FPU_DP" # Double-Precision floating point
    FPU_HP = "FPU_HP" # Half-Precision floating point
    FPU_V4 = "FPUv4" # FPv4, only present in Cortex-M4F
    FPU_V5 = "FPUv5" # FPv5 single or double precision
    SEC = "SEC" # Security Extension
    SEC_V81 = "SEC_V81" # v8.1-M additions to the Security Extension
    MVE = "MVE" # M-profile Vector Extension, with integer support
    MVE_FP = "MVE_FP" # M-profile Vector Extension single- and half-precision floating-point
    UDE = "UDE" # Unprivileged Debug Extension
    RAS = "RAS" # Reliability, Serviceability, and Availability
    PMU = "PMU" # Performance Monitoring Unit
    MPU = "MPU" # Memory Protection Unit
    PACBTI = "PACBTI" # Pointer Authentication and Branch Target Identification



# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import struct
import threading
from enum import Enum
from typing import (List, Optional, Sequence, Tuple, Union)
import usb.core

from .constants import (Commands, Status, SWD_FREQ_MAP, JTAG_FREQ_MAP)
from ...core import exceptions
from ...coresight import dap
from ...utility import conversion
from ...utility.mask import bfx

LOG = logging.getLogger(__name__)

class STLink(object):
    """@brief STLink V2 and V3 command-level interface.
    """
    class Protocol(Enum):
        """@brief Protocol options to pass to STLink.enter_debug() method.
        """
        SWD = 1
        JTAG = 2

    ## Maximum number of bytes to send or receive for 32- and 16- bit transfers.
    #
    # 8-bit transfers have a maximum size of the maximum USB packet size (64 bytes for full speed, 512 for HS).
    MAXIMUM_TRANSFER_SIZE = 6144

    ## Minimum required STLink firmware version (hw version 2).
    MIN_JTAG_VERSION = 24

    ## Firmware version that adds 16-bit transfers (hw version 2).
    MIN_JTAG_VERSION_16BIT_XFER = 26

    ## Firmware version that adds multiple AP support (hw version 2).
    MIN_JTAG_VERSION_MULTI_AP = 28

    ## Firmware version that adds DP bank support.
    #
    # Keys are the hardware version, value is the minimum JTAG version.
    MIN_JTAG_VERSION_DPBANKSEL = {2: 32, 3: 2}

    ## Firmware version that supports JTAG_GET_BOARD_IDENTIFIERS.
    #
    # Keys are the hardware version, value is the minimum JTAG version.
    MIN_JTAG_VERSION_GET_BOARD_IDS = {2: 36, 3: 6}

    ## Firmware versions that support CSW settings on memory access commands.
    #
    # Keys are the hardware version, value is the minimum JTAG version.
    MIN_JTAG_VERSION_MEM_CSW = {2: 32, 3: 2}

    ## Port number to use to indicate DP registers.
    DP_PORT = 0xffff

    ## Map to convert from STLink error response codes to exception classes.
    _ERROR_CLASSES = {
        # AP protocol errors
        Status.SWD_AP_WAIT: exceptions.TransferTimeoutError,
        Status.SWD_AP_FAULT: exceptions.TransferFaultError,
        Status.SWD_AP_ERROR: exceptions.TransferError,
        Status.SWD_AP_PARITY_ERROR: exceptions.TransferError,

        # DP protocol errors
        Status.SWD_DP_WAIT: exceptions.TransferTimeoutError,
        Status.SWD_DP_FAULT: exceptions.TransferFaultError,
        Status.SWD_DP_ERROR: exceptions.TransferError,
        Status.SWD_DP_PARITY_ERROR: exceptions.TransferError,

        # High level transaction errors
        Status.SWD_AP_WDATA_ERROR: exceptions.TransferFaultError,
        Status.SWD_AP_STICKY_ERROR: exceptions.TransferError,
        Status.SWD_AP_STICKYORUN_ERROR: exceptions.TransferError,
        }

    ## These errors indicate a memory fault.
    _MEM_FAULT_ERRORS = (
        Status.JTAG_UNKNOWN_ERROR, # Returned in some cases by older STLink firmware.
        Status.SWD_AP_FAULT,
        Status.SWD_DP_FAULT,
        Status.SWD_AP_WDATA_ERROR,
        Status.SWD_AP_STICKY_ERROR,
        )

    def __init__(self, device):
        self._device = device
        self._hw_version = 0
        self._jtag_version = 0
        self._version_str = None
        self._target_voltage = 0
        self._protocol = None
        self._lock = threading.RLock()

    def open(self):
        with self._lock:
            self._device.open()
            self.enter_idle()
            self.get_version()
            self.get_target_voltage()

    def close(self):
        with self._lock:
            self.enter_idle()
            self._device.close()

    def get_board_id(self) -> Optional[str]:
        """@brief Return the Mbed board ID by command.

        If the device is not already open, it will be temporarily opened in order to read the board ID.

        @retval Board ID as a 4-character string.
        @retval None is returned if the board ID cannot be read.
        """
        with self._lock:
            did_open = False
            try:
                if not self._device.is_open:
                    self.open()
                    did_open = True
                if self._jtag_version < self.MIN_JTAG_VERSION_GET_BOARD_IDS[self._hw_version]:
                    return None
                response = self._device.transfer([Commands.JTAG_COMMAND, Commands.JTAG_GET_BOARD_IDENTIFIERS],
                        readSize=128)
                self._check_status(response[:2])

                # Extract and return the board ID. If the ID field consists of all 0 bytes then we didn't
                # get a valid ID, so return None instead.
                board_id = response[2:6]
                if board_id == b'\x00\x00\x00\x00':
                    return None
                return board_id.decode('ascii')
            except usb.core.USBError:
                return None
            finally:
                if did_open:
                    self.close()

    def get_version(self):
        # GET_VERSION response structure:
        #   Byte 0-1:
        #     [15:12] Major/HW version
        #     [11:6]  JTAG/SWD version
        #     [5:0]   SWIM or MSC version
        #   Byte 2-3: ST_VID
        #   Byte 4-5: STLINK_PID
        response = self._device.transfer([Commands.GET_VERSION], readSize=6)
        ver, = struct.unpack('>H', response[:2])
        # TODO create version bitfield constants
        self._hw_version = bfx(ver, 15, 12)
        self._jtag_version = bfx(ver, 11, 6)
        self._msc_version = bfx(ver, 5, 0)

        # For STLinkV3 we must use the extended get version command.
        if self._hw_version >= 3:
            # GET_VERSION_EXT response structure (byte offsets):
            #   0: HW version
            #   1: SWIM version
            #   2: JTAG/SWD version
            #   3: MSC/VCP version
            #   4: Bridge version
            #   5-7: reserved
            #   8-9: ST_VID
            #   10-11: STLINK_PID
            response = self._device.transfer([Commands.GET_VERSION_EXT], readSize=12)
            hw_vers, _, self._jtag_version, self._msc_version = struct.unpack('<4B', response[0:4])

        self._version_str = "V%dJ%dM%d" % (self._hw_version, self._jtag_version, self._msc_version)
        LOG.debug("STLink probe %s firmware version: %s", self.serial_number, self._version_str)

        # Check versions.
        if self._jtag_version == 0:
            raise exceptions.ProbeError(f"{self._version_str} firmware does not support JTAG/SWD. Please update"
                "to a firmware version that supports JTAG/SWD.")
        if not self._check_version(self.MIN_JTAG_VERSION):
            raise exceptions.ProbeError(f"STLink {self.serial_number} is using an unsupported, older firmware version. "
                f"Please update to the latest STLink firmware. Current version is {self._version_str}, must be at "
                f"least version v2J{self.MIN_JTAG_VERSION}.")

    def _check_version(self, min_version):
        return (self._hw_version >= 3) or (self._jtag_version >= min_version)

    @property
    def vendor_name(self):
        return self._device.vendor_name

    @property
    def product_name(self):
        return self._device.product_name

    @property
    def serial_number(self):
        return self._device.serial_number

    @property
    def hw_version(self):
        return self._hw_version

    @property
    def jtag_version(self):
        return self._jtag_version

    @property
    def version_str(self):
        return self._version_str

    @property
    def target_voltage(self):
        return self._target_voltage

    @property
    def supports_banked_dp(self):
        """@brief Whether the firmware version supports accessing banked DP registers.

        This property is not valid until the connection is opened.
        """
        return self._jtag_version >= self.MIN_JTAG_VERSION_DPBANKSEL[self._hw_version]

    def get_target_voltage(self):
        response = self._device.transfer([Commands.GET_TARGET_VOLTAGE], readSize=8)
        a0, a1 = struct.unpack('<II', response[:8])
        self._target_voltage = 2 * a1 * 1.2 / a0 if a0 != 0 else None

    def enter_idle(self):
        with self._lock:
            response = self._device.transfer([Commands.GET_CURRENT_MODE], readSize=2)
            if response[0] == Commands.DEV_DFU_MODE:
                self._device.transfer([Commands.DFU_COMMAND, Commands.DFU_EXIT])
            elif response[0] == Commands.DEV_JTAG_MODE:
                self._device.transfer([Commands.JTAG_COMMAND, Commands.JTAG_EXIT])
            elif response[0] == Commands.DEV_SWIM_MODE:
                self._device.transfer([Commands.SWIM_COMMAND, Commands.SWIM_EXIT])
            self._protocol = None

    def set_prescaler(self, prescaler: int) -> None:
        assert prescaler in (1, 2, 4)

        # The SWITCH_STLINK_FREQ command is only supported on V3.
        if self._hw_version < 3:
            return
        with self._lock:
            cmd = [Commands.JTAG_COMMAND, Commands.SWITCH_STLINK_FREQ, prescaler]
            response = self._device.transfer(cmd, readSize=8)
            # The JTAG_CONF_CHANGED status is ok and expected.
            if response[0] != Status.JTAG_CONF_CHANGED:
                self._check_status(response[0:2])

    def set_swd_frequency(self, freq: Union[int, float] = 1800000):
        with self._lock:
            if self._hw_version >= 3:
                self.set_com_frequency(self.Protocol.SWD, freq)
            else:
                for f, d in SWD_FREQ_MAP.items():
                    if freq >= f:
                        response = self._device.transfer([Commands.JTAG_COMMAND, Commands.SWD_SET_FREQ, d], readSize=2)
                        self._check_status(response)
                        return
                raise exceptions.ProbeError("Selected SWD frequency is too low")

    def set_jtag_frequency(self, freq: Union[int, float] = 1120000):
        with self._lock:
            if self._hw_version >= 3:
                self.set_com_frequency(self.Protocol.JTAG, freq)
            else:
                for f, d in JTAG_FREQ_MAP.items():
                    if freq >= f:
                        response = self._device.transfer([Commands.JTAG_COMMAND, Commands.JTAG_SET_FREQ, d], readSize=2)
                        self._check_status(response)
                        return
                raise exceptions.ProbeError("Selected JTAG frequency is too low")

    def get_com_frequencies(self, protocol: "STLink.Protocol"):
        assert self._hw_version >= 3

        with self._lock:
            cmd = [Commands.JTAG_COMMAND, Commands.GET_COM_FREQ, protocol.value - 1]
            response = self._device.transfer(cmd, readSize=52)
            self._check_status(response[0:2])

            freqs = list(conversion.byte_list_to_u32le_list(response[4:52]))
            currentFreq = freqs.pop(0)
            freqCount = freqs.pop(0)
            return currentFreq, freqs[:freqCount]

    def set_com_frequency(self, protocol: "STLink.Protocol", freq: Union[int, float]):
        assert self._hw_version >= 3

        with self._lock:
            freq_khz = int(freq) // 1000
            cmd = [Commands.JTAG_COMMAND, Commands.SET_COM_FREQ, protocol.value - 1, 0]
            cmd.extend(conversion.u32le_list_to_byte_list([freq_khz]))
            response = self._device.transfer(cmd, readSize=8)
            self._check_status(response[0:2])

            freqs = conversion.byte_list_to_u32le_list(response[4:8])
            LOG.debug("actual %s frequency is %d kHz", protocol.name, freqs[0])
            return freqs[0]

    def enter_debug(self, protocol):
        with self._lock:
            self.enter_idle()

            if protocol == self.Protocol.SWD:
                protocolParam = Commands.JTAG_ENTER_SWD
            elif protocol == self.Protocol.JTAG:
                protocolParam = Commands.JTAG_ENTER_JTAG_NO_CORE_RESET
            else:
                raise ValueError(protocol)
            response = self._device.transfer([Commands.JTAG_COMMAND, Commands.JTAG_ENTER2, protocolParam, 0], readSize=2)
            self._check_status(response)
            self._protocol = protocol

    def open_ap(self, apsel):
        with self._lock:
            if not self._check_version(self.MIN_JTAG_VERSION_MULTI_AP):
                return
            cmd = [Commands.JTAG_COMMAND, Commands.JTAG_INIT_AP, apsel, Commands.JTAG_AP_NO_CORE]
            response = self._device.transfer(cmd, readSize=2)
            self._check_status(response)

    def close_ap(self, apsel):
        with self._lock:
            if not self._check_version(self.MIN_JTAG_VERSION_MULTI_AP):
                return
            cmd = [Commands.JTAG_COMMAND, Commands.JTAG_CLOSE_AP_DBG, apsel]
            response = self._device.transfer(cmd, readSize=2)
            self._check_status(response)

    def target_reset(self):
        with self._lock:
            response = self._device.transfer([Commands.JTAG_COMMAND, Commands.JTAG_DRIVE_NRST, Commands.JTAG_DRIVE_NRST_PULSE], readSize=2)
            self._check_status(response)

    def drive_nreset(self, isAsserted):
        with self._lock:
            value = Commands.JTAG_DRIVE_NRST_LOW if isAsserted else Commands.JTAG_DRIVE_NRST_HIGH
            response = self._device.transfer([Commands.JTAG_COMMAND, Commands.JTAG_DRIVE_NRST, value], readSize=2)
            self._check_status(response)

    def _check_status(self, response):
        status, = struct.unpack('<H', response)

        if status != Status.JTAG_OK:
            error_message = Status.get_error_message(status)
            if status in self._ERROR_CLASSES:
                raise self._ERROR_CLASSES[status](error_message)
            else:
                raise exceptions.ProbeError(error_message)

    def _clear_sticky_error(self):
        with self._lock:
            if self._protocol == self.Protocol.SWD:
                self.write_dap_register(self.DP_PORT, dap.DP_ABORT,
                    dap.ABORT_ORUNERRCLR | dap.ABORT_WDERRCLR | dap.ABORT_STKERRCLR | dap.ABORT_STKCMPCLR)
            elif self._protocol == self.Protocol.JTAG:
                self.write_dap_register(self.DP_PORT, dap.DP_CTRL_STAT,
                    dap.CTRLSTAT_STICKYERR | dap.CTRLSTAT_STICKYCMP | dap.CTRLSTAT_STICKYORUN)

    def _get_csw_bytes(self, csw: int) -> Tuple[int, int, int]:
        """@brief Return the 3 bytes in little endian order for CSW[31:8] sent in the mem command.
        @param csw CSW to use for the MEM-AP. Only the top 24 bits are used.
        """
        if self._check_version(self.MIN_JTAG_VERSION_MEM_CSW[self._hw_version]):
            return ((csw >> 8) & 0xff), ((csw >> 16) & 0xff), ((csw >> 24) & 0xff)
        else:
            # This version of STLink firmware doesn't support nonstandard CSW.
            # TODO should we log a warning here or elsewhere if csw is set to other than
            # Secure,Priv,Noncacheable,Nonbufferable,Data?
            return 0, 0, 0

    def _read_mem(self, addr: int, size: int, memcmd: int, maxrx: int, apsel: int, csw: int) -> List[int]:
        with self._lock:
            result = []
            while size:
                thisTransferSize = min(size, maxrx)

                # Read Memory {8,16,32} command bytes
                #   0:      JTAG_COMMAND
                #   1:      JTAG_READMEM_{8,16,32}BIT
                #   2-5:    address
                #   6-7:    length in bytes <= 6144 (except 8-bit must <= USB packet size)
                #   8:      APSEL
                #   9-11:   CSW[31:8]
                #   12-15:  TCP unique ID (not used by pyocd)
                cmd = [Commands.JTAG_COMMAND, memcmd]
                cmd.extend(struct.pack('<IHBBBB', addr, thisTransferSize, apsel, *self._get_csw_bytes(csw)))
                result += self._device.transfer(cmd, readSize=thisTransferSize)

                addr += thisTransferSize
                size -= thisTransferSize

                # Check status of this read.
                response = self._device.transfer([Commands.JTAG_COMMAND, Commands.JTAG_GETLASTRWSTATUS2], readSize=12)
                status, _, faultAddr = struct.unpack('<HHI', response[0:8])

                # Handle transfer faults specially so we can assign the address info.
                if status != Status.JTAG_OK:
                    error_message = Status.get_error_message(status)
                    if status in self._MEM_FAULT_ERRORS:
                        # Clear sticky errors.
                        self._clear_sticky_error()

                        exc = exceptions.TransferFaultError("read")
                        exc.fault_address = faultAddr
                        exc.fault_length = thisTransferSize - (faultAddr - addr)
                        raise exc
                    elif status in self._ERROR_CLASSES:
                        raise self._ERROR_CLASSES[status](error_message)
                    elif status != Status.JTAG_OK:
                        raise exceptions.ProbeError(error_message)
            return result

    def _write_mem(self, addr: int, data: Sequence[int], memcmd: int, maxtx: int, apsel: int, csw: int) -> None:
        with self._lock:
            while len(data):
                thisTransferSize = min(len(data), maxtx)
                thisTransferData = data[:thisTransferSize]

                # Write Memory {8,16,32} command bytes
                #   0:      JTAG_COMMAND
                #   1:      JTAG_WRITEMEM_{8,16,32}BIT
                #   2-5:    address
                #   6-7:    length in bytes <= 6144 (except 8-bit must <= USB packet size)
                #   8:      APSEL
                #   9-11:   CSW[31:8]
                #   12-15:  TCP unique ID (not used by pyocd)
                cmd = [Commands.JTAG_COMMAND, memcmd]
                cmd.extend(struct.pack('<IHBBBB', addr, thisTransferSize, apsel, *self._get_csw_bytes(csw)))
                self._device.transfer(cmd, writeData=thisTransferData)

                addr += thisTransferSize
                data = data[thisTransferSize:]

                # Check status of this write.
                response = self._device.transfer([Commands.JTAG_COMMAND, Commands.JTAG_GETLASTRWSTATUS2], readSize=12)
                status, _, faultAddr = struct.unpack('<HHI', response[0:8])

                # Handle transfer faults specially so we can assign the address info.
                if status != Status.JTAG_OK:
                    error_message = Status.get_error_message(status)
                    if status in self._MEM_FAULT_ERRORS:
                        # Clear sticky errors.
                        self._clear_sticky_error()

                        exc = exceptions.TransferFaultError("write")
                        exc.fault_address = faultAddr
                        exc.fault_length = thisTransferSize - (faultAddr - addr)
                        raise exc
                    elif status in self._ERROR_CLASSES:
                        raise self._ERROR_CLASSES[status](error_message)
                    elif status != Status.JTAG_OK:
                        raise exceptions.ProbeError(error_message)

    def read_mem32(self, addr: int, size: int, apsel: int, csw: int):
        assert (addr & 0x3) == 0 and (size & 0x3) == 0, "address and size must be word aligned"
        return self._read_mem(addr, size, Commands.JTAG_READMEM_32BIT, self.MAXIMUM_TRANSFER_SIZE, apsel, csw)

    def write_mem32(self, addr: int, data: Sequence[int], apsel: int, csw: int):
        assert (addr & 0x3) == 0 and (len(data) & 3) == 0, "address and size must be word aligned"
        self._write_mem(addr, data, Commands.JTAG_WRITEMEM_32BIT, self.MAXIMUM_TRANSFER_SIZE, apsel, csw)

    def read_mem16(self, addr: int, size: int, apsel: int, csw: int):
        assert (addr & 0x1) == 0 and (size & 0x1) == 0, "address and size must be half-word aligned"

        if not self._check_version(self.MIN_JTAG_VERSION_16BIT_XFER):
            # 16-bit r/w is only available from J26, so revert to 8-bit accesses.
            return self.read_mem8(addr, size, apsel, csw)

        return self._read_mem(addr, size, Commands.JTAG_READMEM_16BIT, self.MAXIMUM_TRANSFER_SIZE, apsel, csw)

    def write_mem16(self, addr: int, data: Sequence[int], apsel: int, csw: int):
        assert (addr & 0x1) == 0 and (len(data) & 1) == 0, "address and size must be half-word aligned"

        if not self._check_version(self.MIN_JTAG_VERSION_16BIT_XFER):
            # 16-bit r/w is only available from J26, so revert to 8-bit accesses.
            self.write_mem8(addr, data, apsel, csw)
            return

        self._write_mem(addr, data, Commands.JTAG_WRITEMEM_16BIT, self.MAXIMUM_TRANSFER_SIZE, apsel, csw)

    def read_mem8(self, addr: int, size: int, apsel: int, csw: int):
        return self._read_mem(addr, size, Commands.JTAG_READMEM_8BIT, self._device.max_packet_size, apsel, csw)

    def write_mem8(self, addr: int, data: Sequence[int], apsel: int, csw: int):
        self._write_mem(addr, data, Commands.JTAG_WRITEMEM_8BIT, self._device.max_packet_size, apsel, csw)

    def _check_dp_bank(self, port, addr):
        """@brief Check if attempting to access a banked DP register with a firmware version that
                doesn't support that.
        """
        if ((port == self.DP_PORT) and ((addr & 0xf0) != 0) and not self.supports_banked_dp):
            raise exceptions.ProbeError(f"this STLinkV{self._hw_version} firmware version does not support accessing"
                    f" banked DP registers; please upgrade to the latest STLinkV{self._hw_version} firmware release")

    def read_dap_register(self, port, addr):
        assert (addr >> 16) == 0, "register address must be 16-bit"

        self._check_dp_bank(port, addr)

        with self._lock:
            cmd = [Commands.JTAG_COMMAND, Commands.JTAG_READ_DAP_REG]
            cmd.extend(struct.pack('<HH', port, addr))
            response = self._device.transfer(cmd, readSize=8)
            self._check_status(response[:2])
            value, = struct.unpack('<I', response[4:8])
            return value

    def write_dap_register(self, port, addr, value):
        assert (addr >> 16) == 0, "register address must be 16-bit"

        self._check_dp_bank(port, addr)

        with self._lock:
            cmd = [Commands.JTAG_COMMAND, Commands.JTAG_WRITE_DAP_REG]
            cmd.extend(struct.pack('<HHI', port, addr, value))
            response = self._device.transfer(cmd, readSize=2)
            self._check_status(response)

    def swo_start(self, baudrate):
        with self._lock:
            bufferSize = 4096
            cmd = [Commands.JTAG_COMMAND, Commands.SWV_START_TRACE_RECEPTION]
            cmd.extend(struct.pack('<HI', bufferSize, baudrate))
            response = self._device.transfer(cmd, readSize=2)
            self._check_status(response)

    def swo_stop(self):
        with self._lock:
            cmd = [Commands.JTAG_COMMAND, Commands.SWV_STOP_TRACE_RECEPTION]
            response = self._device.transfer(cmd, readSize=2)
            self._check_status(response)

    def swo_read(self):
        with self._lock:
            response = None
            bytesAvailable = None
            try:
                cmd = [Commands.JTAG_COMMAND, Commands.SWV_GET_TRACE_NEW_RECORD_NB]
                response = self._device.transfer(cmd, readSize=2)
                bytesAvailable, = struct.unpack('<H', response)
                if bytesAvailable:
                    return self._device.read_swv(bytesAvailable)
                else:
                    return bytearray()
            except KeyboardInterrupt:
                # If we're interrupted after sending the SWV_GET_TRACE_NEW_RECORD_NB command,
                # we have to read the queued SWV data before any other commands can be sent.
                if response is not None:
                    if bytesAvailable is None:
                        bytesAvailable, = struct.unpack('<H', response)
                    if bytesAvailable:
                        self._device.read_swv(bytesAvailable)



# pyOCD debugger
# Copyright (c) 2006-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# Copyright (c) 2022 Clay McClure
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
from time import sleep
import sys
import io
from xml.etree.ElementTree import (Element, SubElement, tostring)
from typing import (Dict, List, Optional, Tuple)

from ..core import exceptions
from ..core.target import Target
from ..flash.loader import FlashLoader
from ..utility.cmdline import convert_vector_catch
from ..utility.conversion import (hex_to_byte_list, hex_encode, hex_decode, hex8_to_u32le)
from ..utility.compatibility import (to_bytes_safe, to_str_safe)
from ..utility.server import StreamServer
from ..utility.timeout import Timeout
from ..trace.swv import SWVReader
from ..utility.rtt_server import RTTServer
from ..utility.sockets import ListenerSocket
from .syscall import GDBSyscallIOHandler
from ..debug import semihost
from .context_facade import GDBDebugContextFacade
from .symbols import GDBSymbolProvider
from ..rtos import RTOS
from . import signals
from .packet_io import (
    checksum,
    ConnectionClosedException,
    GDBServerPacketIOThread,
    )
from ..commands.execution_context import CommandExecutionContext
from ..commands.commander import ToolExitException

# Import this module, even though it's not used below, to ensure the gdbserver commands get loaded.
from . import gdbserver_commands # noqa

LOG = logging.getLogger(__name__)

TRACE_MEM = LOG.getChild("trace.mem")
TRACE_MEM.setLevel(logging.CRITICAL)

def unescape(data: bytes) -> List[int]:
    """@brief De-escapes binary data from Gdb.

    @param data Bytes-like object with possibly escaped values.
    @return List of integers in the range 0-255, with all escaped bytes de-escaped.
    """
    data_idx = 0

    # unpack the data into binary array
    result = list(data)

    # check for escaped characters
    while data_idx < len(result):
        if result[data_idx] == 0x7d:
            result.pop(data_idx)
            result[data_idx] = result[data_idx] ^ 0x20
        data_idx += 1

    return result

## Tuple of int values of characters that must be escaped.
_GDB_ESCAPED_CHARS = tuple(b'#$}*')

def escape(data):
    """@brief Escape binary data to be sent to Gdb.

    @param data Bytes-like object containing raw binary.
    @return Bytes object with the characters in '#$}*' escaped as required by Gdb.
    """
    result: List[int] = []
    for c in data:
        if c in _GDB_ESCAPED_CHARS:
            # Escape by prefixing with '}' and xor'ing the char with 0x20.
            result += [0x7d, c ^ 0x20]
        else:
            result.append(c)
    return bytes(result)

class GDBServer(threading.Thread):
    """@brief GDB remote server thread.

    This class start a GDB server listening a gdb connection on a specific port.
    It implements the RSP (Remote Serial Protocol).
    """

    ## Notification event for the gdbserver beginnning to listen on its RSP port.
    GDBSERVER_START_LISTENING_EVENT = 'gdbserver-start-listening'

    ## Timer delay for sending the notification that the server is listening.
    START_LISTENING_NOTIFY_DELAY = 0.03 # 30 ms

    def __init__(self, session, core=None):
        super().__init__()
        self.session = session
        self.board = session.board
        if core is None:
            self.core = 0
            self.target = self.board.target
        else:
            self.core = core
            self.target = self.board.target.cores[core]
        self.name = "gdb-server-core%d" % self.core

        self.port = session.options.get('gdbserver_port')
        if self.port != 0:
            self.port += self.core
        self.telnet_port = session.options.get('telnet_port')
        if self.telnet_port != 0:
            self.telnet_port += self.core

        self.vector_catch = session.options.get('vector_catch')
        self.target.set_vector_catch(convert_vector_catch(self.vector_catch))
        self.step_into_interrupt = session.options.get('step_into_interrupt')
        self.persist = session.options.get('persist')
        self.enable_semihosting = session.options.get('enable_semihosting')
        self.semihost_console_type = session.options.get('semihost_console_type') # Not subscribed.
        self.semihost_use_syscalls = session.options.get('semihost_use_syscalls') # Not subscribed.
        self.serve_local_only = session.options.get('serve_local_only') # Not subscribed.
        self.report_core = session.options.get('report_core_number')
        # Subscribe to changes for those of the above options that make sense to change at runtime.
        self.session.options.subscribe(self._option_did_change, [
                'vector_catch',
                'step_into_interrupt',
                'persist',
                'enable_semihosting',
                'report_core_number',
                ])

        self.packet_size = 2048
        self.packet_io = None
        self.gdb_features = []
        self.non_stop = False
        self._is_extended_remote = False
        self.is_target_running = (self.target.get_state() == Target.State.RUNNING)
        self.flash_loader = None
        self.shutdown_event = threading.Event()
        self.detach_event = threading.Event()
        if core is None:
            self.target_context = self.board.target.get_target_context()
        else:
            self.target_context = self.board.target.get_target_context(core=core)
        self.target_facade = GDBDebugContextFacade(self.target_context)
        self.thread_provider = None
        self.did_init_thread_providers = False
        self.current_thread_id = 0
        self.first_run_after_reset_or_flash = True

        self.abstract_socket = ListenerSocket(self.port, self.packet_size)
        if not self.serve_local_only:
            # We really should be binding to explicit interfaces, not all available.
            self.abstract_socket.host = ''
        self.abstract_socket.init()
        # Read back bound port in case auto-assigned (port 0)
        self.port = self.abstract_socket.port

        # Coarse grain lock to synchronize SWO with other activity
        self.lock = threading.Lock()

        self.session.subscribe(self.event_handler, Target.Event.POST_RESET)

        # Init semihosting and telnet console.
        if self.semihost_use_syscalls:
            semihost_io_handler = GDBSyscallIOHandler(self)
        else:
            # Use internal IO handler.
            semihost_io_handler = semihost.InternalSemihostIOHandler()

        if self.semihost_console_type == 'telnet':
            self.telnet_server = StreamServer(self.telnet_port, self.serve_local_only, "Semihost",
                False, extra_info=("core %d" % self.core))
            console_file = self.telnet_server
            semihost_console = semihost.ConsoleIOHandler(self.telnet_server)
        else:
            LOG.info("Semihosting will be output to console")
            console_file = sys.stdout
            self.telnet_server = None
            semihost_console = semihost_io_handler
        self.semihost = semihost.SemihostAgent(self.target_context, io_handler=semihost_io_handler, console=semihost_console)

        # Start with RTT disabled
        self.rtt_server: Optional[RTTServer] = None

        #
        # If SWV is enabled, create a SWVReader thread. Note that we only do
        # this if the core is 0: SWV is not a per-core construct, and can't
        # be meaningfully read by multiple threads concurrently.
        #
        self._swv_reader = None

        if session.options.get("enable_swv") and core == 0:
            if "swv_system_clock" not in session.options:
                LOG.warning("Cannot enable SWV due to missing swv_system_clock option")
            else:
                sys_clock = int(session.options.get("swv_system_clock"))
                swo_clock = int(session.options.get("swv_clock"))
                self._swv_reader = SWVReader(session, self.core, self.lock)
                self._swv_reader.init(sys_clock, swo_clock, console_file)

        self._init_remote_commands()

        # pylint: disable=invalid-name

        # Command handler table.
        #
        # The dict keys are the first character of the incoming command from gdb. Values are a
        # bi-tuple. The first element is the handler method, and the second element is the start
        # offset of the command string passed to the handler.
        #
        # Start offset values:
        #  0 - Special case: handler method does not take any parameters.
        #  1 - Strip off leading "$" from command.
        #  2 - Strip leading "$" plus character matched through this table.
        #  3+ - Supported, but not very useful.
        #
        self.COMMANDS = {
        #       CMD    HANDLER                   START    DESCRIPTION
                b'!' : (self.extended_remote,    0   ), # Enable extended remote mode.
                b'?' : (self.stop_reason_query,  0   ), # Stop reason query.
                b'c' : (self.resume,             1   ), # Continue (at addr)
                b'C' : (self.resume,             1   ), # Continue with signal.
                b'D' : (self.detach,             1   ), # Detach.
                b'g' : (self.get_registers,      0   ), # Read general registers.
                b'G' : (self.set_registers,      2   ), # Write general registers.
                b'H' : (self.set_thread,         2   ), # Set thread for subsequent operations.
                b'k' : (self.kill,               0   ), # Kill.
                b'm' : (self.get_memory,         2   ), # Read memory.
                b'M' : (self.write_memory_hex,   2   ), # Write memory (hex).
                b'p' : (self.read_register,      2   ), # Read register.
                b'P' : (self.write_register,     2   ), # Write register.
                b'q' : (self.handle_query,       2   ), # General query.
                b'Q' : (self.handle_general_set, 2   ), # General set.
                b'R' : (self.restart,            1   ), # Extended remote restart command.
                b's' : (self.step,               1   ), # Single step.
                b'S' : (self.step,               1   ), # Step with signal.
                b'T' : (self.is_thread_alive,    1   ), # Thread liveness query.
                b'v' : (self.v_command,          2   ), # v command.
                b'X' : (self.write_memory,       2   ), # Write memory (binary).
                b'z' : (self.breakpoint,         1   ), # Insert breakpoint/watchpoint.
                b'Z' : (self.breakpoint,         1   ), # Remove breakpoint/watchpoint.
            }

        # pylint: enable=invalid-name

        self.setDaemon(True)

    def _init_remote_commands(self):
        """@brief Initialize the remote command processor infrastructure."""
        # Create command execution context. The output stream will default to stdout
        # but we'll change it to a fresh StringIO prior to running each command.
        #
        # Note we also modify the selected_core property so it is initially set to the gdbserver's core.
        self._command_context = CommandExecutionContext()
        self._command_context.selected_core = self.target
        self._command_context.attach_session(self.session)

        # Add the gdbserver command group.
        self._command_context.command_set.add_command_group('gdbserver')

    def stop(self, wait=True):
        if self.is_alive():
            self.shutdown_event.set()
            if wait:
                LOG.debug("gdbserver shutdown event set; waiting for exit")
                self.join()
            else:
                LOG.debug("gdbserver shutdown event set")
            LOG.info("GDB server thread stopped")

    def _cleanup(self):
        LOG.debug("GDB server cleaning up")
        if self.packet_io:
            self.packet_io.stop()
            self.packet_io = None
        if self.semihost:
            self.semihost.cleanup()
            self.semihost = None
        if self.telnet_server:
            self.telnet_server.stop()
            self.telnet_server = None
        if self._swv_reader:
            self._swv_reader.stop()
            self._swv_reader = None
        if self.rtt_server:
            self.rtt_server.stop()
            self.rtt_server = None
        self.abstract_socket.cleanup()

    def _cleanup_for_next_connection(self):
        self.non_stop = False
        self.thread_provider = None
        self.did_init_thread_providers = False
        self.current_thread_id = 0

    def run(self):
        LOG.info('GDB server started on port %d (core %d)', self.port, self.core)

        while not self.shutdown_event.is_set():
            try:
                # Notify listeners that the server is running after a short delay.
                #
                # This timer prevents a race condition where the notification is sent before the server is
                # actually listening. It's not a 100% guarantee, though.
                notify_timer = threading.Timer(self.START_LISTENING_NOTIFY_DELAY, self.session.notify,
                        args=(self.GDBSERVER_START_LISTENING_EVENT, self))
                notify_timer.start()

                while not self.shutdown_event.is_set():
                    connected = self.abstract_socket.connect()
                    if connected != None:
                        self.packet_io = GDBServerPacketIOThread(self.abstract_socket)
                        break

                if self.shutdown_event.is_set():
                    break

                # Make sure the target is halted. Otherwise gdb gets easily confused.
                self.target.halt()

                LOG.info("Client connected to port %d!", self.port)
                self._run_connection()
                LOG.info("Client disconnected from port %d!", self.port)

            except Exception as e:
                LOG.error("Unexpected exception: %s", e, exc_info=self.session.log_tracebacks)

        LOG.debug("gdbserver thread exiting")
        self._cleanup()

    def _run_connection(self):
        assert self.packet_io

        self.detach_event.clear()

        while not (self.detach_event.is_set() or self.shutdown_event.is_set()):
            try:
                if self.packet_io.interrupt_event.is_set():
                    if self.non_stop:
                        self.target.halt()
                        self.is_target_running = False
                        self.send_stop_notification()
                    else:
                        LOG.warning("Got unexpected ctrl-c, ignoring")
                    self.packet_io.interrupt_event.clear()

                if self.non_stop and self.is_target_running:
                    try:
                        if self.target.get_state() == Target.State.HALTED:
                            LOG.debug("state halted")
                            self.is_target_running = False
                            self.send_stop_notification()
                    except Exception as e:
                        LOG.error("Unexpected exception: %s", e, exc_info=self.session.log_tracebacks)

                # read command
                try:
                    packet = self.packet_io.receive(block=not self.non_stop)
                except ConnectionClosedException:
                    LOG.debug("gdbserver connection loop exiting; client closed connection")
                    break

                if self.shutdown_event.is_set():
                    break

                if self.non_stop and packet is None:
                    sleep(0.1)
                    continue

                if packet is not None and len(packet) != 0:
                    # decode and prepare resp
                    resp = self.handle_message(packet)

                    if resp is not None:
                        # send resp
                        self.packet_io.send(resp)

            except Exception as e:
                LOG.error("Unexpected exception: %s", e, exc_info=self.session.log_tracebacks)

        LOG.debug("gdbserver exiting connection loop")

        # Clean up the connection.
        self.abstract_socket.close()
        self.packet_io.stop()
        self.packet_io = None

        # If persisting is not enabled, we exit on detach. Otherwise prepare for a new connection.
        if self.persist:
            LOG.debug("preparing for next connection")
            self._cleanup_for_next_connection()
        else:
            self.shutdown_event.set()

    def handle_message(self, msg):
        try:
            assert msg[0:1] == b'$', "invalid first char of message != $"

            try:
                handler, msgStart = self.COMMANDS[msg[1:2]]
            except (KeyError, IndexError):
                LOG.error("Unknown RSP command (%s)", msg[1:2])
                return self.create_rsp_packet(b"")

            with self.lock:
                if msgStart == 0:
                    reply = handler()
                else:
                    reply = handler(msg[msgStart:])

            return reply

        except Exception as e:
            LOG.error("Unhandled exception in handle_message (%s): %s",
                    msg[1:2], e, exc_info=self.session.log_tracebacks)
            return self.create_rsp_packet(b"E01")

    def extended_remote(self):
        LOG.debug("extended remote enabled")
        self._is_extended_remote = True
        return self.create_rsp_packet(b"OK")

    def detach(self, data):
        LOG.info("Client detached")
        # In extended-remote mode, detach should detach from the program but not close the connection. gdb assumes
        # the server connection is still valid. Detaching from the program doesn't really make sense for embedded
        # targets, so just ignore the detach.
        if not self._is_extended_remote:
            self.detach_event.set()
        return self.create_rsp_packet(b"OK")

    def kill(self):
        LOG.debug("GDB kill")
        self.detach_event.set()
        # No packet is returned from the 'k' command.

    def restart(self, data):
        self.target.reset_and_halt()
        # No reply.

    def breakpoint(self, data):
        # handle breakpoint/watchpoint commands
        split = data.split(b'#')[0].split(b',')
        addr = int(split[1], 16)
        LOG.debug("GDB breakpoint %s%d @ %x" % (data[0:1], int(data[1:2]), addr))

        # handle software breakpoint Z0/z0
        if data[1:2] == b'0':
            if data[0:1] == b'Z':
                if not self.target.set_breakpoint(addr, Target.BreakpointType.SW):
                    return self.create_rsp_packet(b'E01') #EPERM
            else:
                self.target.remove_breakpoint(addr)
            return self.create_rsp_packet(b"OK")

        # handle hardware breakpoint Z1/z1
        if data[1:2] == b'1':
            if data[0:1] == b'Z':
                if self.target.set_breakpoint(addr, Target.BreakpointType.HW) is False:
                    return self.create_rsp_packet(b'E01') #EPERM
            else:
                self.target.remove_breakpoint(addr)
            return self.create_rsp_packet(b"OK")

        # handle hardware watchpoint Z2/z2/Z3/z3/Z4/z4
        if data[1:2] == b'2':
            # Write-only watch
            watchpoint_type = Target.WatchpointType.WRITE
        elif data[1:2] == b'3':
            # Read-only watch
            watchpoint_type = Target.WatchpointType.READ
        elif data[1:2] == b'4':
            # Read-Write watch
            watchpoint_type = Target.WatchpointType.READ_WRITE
        else:
            return self.create_rsp_packet(b'E01') #EPERM

        size = int(split[2], 16)
        if data[0:1] == b'Z':
            if self.target.set_watchpoint(addr, size, watchpoint_type) is False:
                return self.create_rsp_packet(b'E01') #EPERM
        else:
            self.target.remove_watchpoint(addr, size, watchpoint_type)
        return self.create_rsp_packet(b"OK")

    def set_thread(self, data):
        if not self.is_threading_enabled():
            return self.create_rsp_packet(b'OK')

        LOG.debug("set_thread:%s", data)
        op = data[0:1]
        thread_id = int(data[1:-3], 16)
        if not (thread_id in (0, -1) or self.thread_provider.is_valid_thread_id(thread_id)):
            return self.create_rsp_packet(b'E01')

        if op == b'c':
            pass
        elif op == b'g':
            if thread_id == -1:
                self.target_facade.set_context(self.target_context)
            else:
                if thread_id == 0:
                    thread = self.thread_provider.current_thread
                    thread_id = thread.unique_id
                else:
                    thread = self.thread_provider.get_thread(thread_id)
                self.target_facade.set_context(thread.context)
        else:
            return self.create_rsp_packet(b'E01')

        self.current_thread_id = thread_id
        return self.create_rsp_packet(b'OK')

    def is_thread_alive(self, data):
        threadId = int(data[1:-3], 16)

        if self.is_threading_enabled():
            isAlive = self.thread_provider.is_valid_thread_id(threadId)
        else:
            isAlive = (threadId == 1)

        if isAlive:
            return self.create_rsp_packet(b'OK')
        else:
            self.validate_debug_context()
            return self.create_rsp_packet(b'E00')

    def validate_debug_context(self):
        if self.is_threading_enabled():
            currentThread = self.thread_provider.current_thread
            if self.current_thread_id != currentThread.unique_id:
                self.target_facade.set_context(currentThread.context)
                self.current_thread_id = currentThread.unique_id
        else:
            if self.current_thread_id != 1:
                LOG.debug("Current thread %x is no longer valid, switching context to target", self.current_thread_id)
                self.target_facade.set_context(self.target_context)
                self.current_thread_id = 1

    def stop_reason_query(self):
        # In non-stop mode, if no threads are stopped we need to reply with OK.
        if self.non_stop and self.is_target_running:
            return self.create_rsp_packet(b"OK")

        return self.create_rsp_packet(self.get_t_response())

    def _get_resume_step_addr(self, data):
        if data is None:
            return None
        data = data.split(b'#')[0]
        if b';' not in data:
            return None
        # c[;addr]
        if data[0:1] in (b'c', b's'):
            addr = int(data[2:], base=16)
        # Csig[;addr]
        elif data[0:1] in (b'C', b'S'):
            addr = int(data[1:].split(b';')[1], base=16)
        else:
            raise exceptions.DebugError("invalid step address received from gdb")
        return addr

    def resume(self, data):
#         addr = self._get_resume_step_addr(data)
        self.target.resume()
        LOG.debug("target resumed")

        if self.first_run_after_reset_or_flash:
            self.first_run_after_reset_or_flash = False
            if self.thread_provider is not None:
                self.thread_provider.read_from_target = True

        val = b''

        # Timeout used only if the target starts returning faults. The is_running property of this timeout
        # also serves as a flag that a fault occurred and we're attempting to retry.
        fault_retry_timeout = Timeout(self.session.options.get('debug.status_fault_retry_timeout'))

        while fault_retry_timeout.check():
            if self.shutdown_event.is_set():
                self.packet_io.interrupt_event.clear()
                return self.create_rsp_packet(val)

            self.lock.release()

            # Wait for a ctrl-c to be received.
            if self.packet_io.interrupt_event.wait(0.01):
                self.lock.acquire()
                LOG.debug("receive CTRL-C")
                self.packet_io.interrupt_event.clear()

                # Be careful about reading the target state. If we previously got a fault (the timeout
                # is running) then ignore the error. In all cases we still return SIGINT.
                try:
                    self.target.halt()
                    val = self.get_t_response(forceSignal=signals.SIGINT)
                except exceptions.TransferError as e:
                    # Note: if the target is not actually halted, gdb can get confused from this point on.
                    # But there's not much we can do if we're getting faults attempting to control it.
                    if not fault_retry_timeout.is_running:
                        LOG.error('Error reading target status: %s', e, exc_info=self.session.log_tracebacks)
                    val = ('S%02x' % signals.SIGINT).encode()
                break

            self.lock.acquire()

            try:
                state = self.target.get_state()

                if self.rtt_server:
                    self.rtt_server.poll()

                # If we were able to successfully read the target state after previously receiving a fault,
                # then clear the timeout.
                if fault_retry_timeout.is_running:
                    LOG.info("Target control reestablished.")
                    fault_retry_timeout.clear()

                if state == Target.State.HALTED:
                    # Handle semihosting
                    if self.enable_semihosting:
                        was_semihost = self.semihost.check_and_handle_semihost_request()

                        if was_semihost:
                            self.target.resume()
                            continue

                    pc = self.target_context.read_core_register('pc')
                    LOG.debug("state halted; pc=0x%08x", pc)
                    val = self.get_t_response()
                    break
            except exceptions.TransferError as e:
                # If we get any sort of transfer error or fault while checking target status, then start
                # a timeout running. Upon a later successful status check, the timeout is cleared. In the event
                # that the timeout expires, this loop is exited and an error raised to gdb.
                if not fault_retry_timeout.is_running:
                    LOG.warning("Transfer error while checking target status; retrying: %s", e,
                            exc_info=self.session.log_tracebacks)
                fault_retry_timeout.start()
            except exceptions.Error as e:
                try:
                    self.target.halt()
                except exceptions.Error:
                    pass
                LOG.warning('Error while target was running: %s', e, exc_info=self.session.log_tracebacks)
                # This exception was not a transfer error, so reading the target state should be ok.
                val = ('S%02x' % self.target_facade.get_signal_value()).encode()
                break

        # Check if we exited the above loop due to a timeout after a fault.
        if fault_retry_timeout.did_time_out:
            LOG.error("Timed out while attempting to reestablish control over target.")
            val = ('S%02x' % signals.SIGSEGV).encode()

        return self.create_rsp_packet(val)

    def step(self, data, start=0, end=0):
        #addr = self._get_resume_step_addr(data)
        LOG.debug("GDB step: %s (start=0x%x, end=0x%x)", data, start, end)

        # Use the step hook to check for an interrupt event.
        def step_hook():
            # Note we don't clear the interrupt event here!
            return self.packet_io.interrupt_event.is_set()
        self.target.step(not self.step_into_interrupt, start, end, hook_cb=step_hook)

        # Clear and handle an interrupt.
        if self.packet_io.interrupt_event.is_set():
            LOG.debug("Received Ctrl-C during step")
            self.packet_io.interrupt_event.clear()
            response = self.get_t_response(forceSignal=signals.SIGINT)
        else:
            response = self.get_t_response()

        return self.create_rsp_packet(response)

    def halt(self):
        self.target.halt()
        return self.create_rsp_packet(self.get_t_response())

    def send_stop_notification(self, forceSignal=None):
        data = self.get_t_response(forceSignal=forceSignal)
        packet = b'%Stop:' + data + b'#' + checksum(data)
        self.packet_io.send(packet)

    def v_command(self, data):
        cmd = data.split(b'#')[0]

        # Flash command.
        if cmd.startswith(b'Flash'):
            return self.flash_op(data)

        # v_cont capabilities query.
        elif b'Cont?' == cmd:
            return self.create_rsp_packet(b"vCont;c;C;s;S;r;t")

        # v_cont, thread action command.
        elif cmd.startswith(b'Cont'):
            return self.v_cont(cmd)

        # vStopped, part of thread stop state notification sequence.
        elif b'Stopped' in cmd:
            # Because we only support one thread for now, we can just reply OK to vStopped.
            return self.create_rsp_packet(b"OK")

        return self.create_rsp_packet(b"")

    # Example: $v_cont;s:1;c#c1
    def v_cont(self, cmd):
        ops = cmd.split(b';')[1:] # split and remove 'Cont' from list
        if not ops:
            return self.create_rsp_packet(b"OK")

        # Maps the thread unique ID to an action char (byte).
        thread_actions: Dict[int, Optional[bytes]] = {}

        if self.is_threading_enabled():
            threads = self.thread_provider.get_threads()
            for k in threads:
                thread_actions[k.unique_id] = None
            currentThread = self.thread_provider.get_current_thread_id()
        else:
            thread_actions[1] = None # our only thread
            currentThread = 1
        default_action = None

        for op in ops:
            args = op.split(b':')
            action = args[0]
            if len(args) > 1:
                thread_id = int(args[1], 16)
                if thread_id == -1 or thread_id == 0:
                    thread_id = currentThread
                thread_actions[thread_id] = action
            else:
                default_action = action

        LOG.debug("thread_actions=%s; default_action=%s", repr(thread_actions), default_action)

        # Only the current thread is supported at the moment.
        if thread_actions[currentThread] is None:
            if default_action is None:
                return self.create_rsp_packet(b'E01')
            thread_actions[currentThread] = default_action

        if thread_actions[currentThread][0:1] in (b'c', b'C'):
            if self.non_stop:
                self.target.resume()
                self.is_target_running = True
                return self.create_rsp_packet(b"OK")
            else:
                return self.resume(None)
        elif thread_actions[currentThread][0:1] in (b's', b'S', b'r'):
            start = 0
            end = 0
            if thread_actions[currentThread][0:1] == b'r':
                start, end = [int(addr, base=16) for addr in thread_actions[currentThread][1:].split(b',')]

            if self.non_stop:
                self.target.step(not self.step_into_interrupt, start, end)
                self.packet_io.send(self.create_rsp_packet(b"OK"))
                self.send_stop_notification()
                return None
            else:
                return self.step(None, start, end)
        elif thread_actions[currentThread] == b't':
            # Must ignore t command in all-stop mode.
            if not self.non_stop:
                return self.create_rsp_packet(b"")
            self.packet_io.send(self.create_rsp_packet(b"OK"))
            self.target.halt()
            self.is_target_running = False
            self.send_stop_notification(forceSignal=0)
        else:
            LOG.error("Unsupported v_cont action '%s'" % thread_actions[1])

    def flash_op(self, data):
        ops = data.split(b':')[0]
        LOG.debug("flash op: %s", ops)

        if ops == b'FlashErase':
            return self.create_rsp_packet(b"OK")

        elif ops == b'FlashWrite':
            write_addr = int(data.split(b':')[1], 16)
            LOG.debug("flash write addr: 0x%x", write_addr)
            # search for second ':' (beginning of data encoded in the message)
            second_colon = 0
            idx_begin = 0
            while second_colon != 2:
                if data[idx_begin:idx_begin+1] == b':':
                    second_colon += 1
                idx_begin += 1

            # Get flash loader if there isn't one already
            if self.flash_loader is None:
                self.flash_loader = FlashLoader(self.session)

            # Add data to flash loader
            self.flash_loader.add_data(write_addr, unescape(data[idx_begin:len(data) - 3]))

            return self.create_rsp_packet(b"OK")

        # we need to flash everything
        elif b'FlashDone' in ops :
            # Only program if we received data.
            if self.flash_loader is not None:
                try:
                    # Write all buffered flash contents.
                    self.flash_loader.commit()
                finally:
                    # Set flash loader to None so that on the next flash command a new
                    # object is used.
                    self.flash_loader = None

            self.first_run_after_reset_or_flash = True
            if self.thread_provider is not None:
                self.thread_provider.read_from_target = False

            return self.create_rsp_packet(b"OK")

        return None

    def get_memory(self, data):
        split = data.split(b',')
        addr = int(split[0], 16)
        length = split[1].split(b'#')[0]
        length = int(length, 16)

        TRACE_MEM.debug("GDB getMem: addr=%x len=%x", addr, length)

        try:
            mem = self.target_context.read_memory_block8(addr, length)
            # Flush so an exception is thrown now if invalid memory was accesses
            self.target_context.flush()
            val = hex_encode(bytearray(mem))
        except exceptions.TransferError as e:
            LOG.debug("get_memory failed at 0x%x: %s", addr, str(e))
            val = b'E01' #EPERM
        return self.create_rsp_packet(val)

    def write_memory_hex(self, data):
        split = data.split(b',')
        addr = int(split[0], 16)

        split = split[1].split(b':')
        length = int(split[0], 16)

        split = split[1].split(b'#')
        data = hex_to_byte_list(split[0])

        TRACE_MEM.debug("GDB writeMemHex: addr=%x len=%x", addr, length)

        try:
            if length > 0:
                self.target_context.write_memory_block8(addr, data)
                # Flush so an exception is thrown now if invalid memory was accessed
                self.target_context.flush()
            resp = b"OK"
        except exceptions.TransferError as e:
            LOG.debug("write_memory_hex failed at 0x%x: %s", addr, str(e))
            resp = b'E01' #EPERM

        return self.create_rsp_packet(resp)

    def write_memory(self, data):
        split = data.split(b',')
        addr = int(split[0], 16)
        length = int(split[1].split(b':')[0], 16)

        TRACE_MEM.debug("GDB writeMem: addr=%x len=%x", addr, length)

        idx_begin = data.index(b':') + 1
        data = data[idx_begin:len(data) - 3]
        data = unescape(data)

        try:
            if length > 0:
                self.target_context.write_memory_block8(addr, data)
                # Flush so an exception is thrown now if invalid memory was accessed
                self.target_context.flush()
            resp = b"OK"
        except exceptions.TransferError as e:
            LOG.debug("write_memory failed at 0x%x: %s", addr, str(e))
            resp = b'E01' #EPERM

        return self.create_rsp_packet(resp)

    def read_register(self, which):
        return self.create_rsp_packet(self.target_facade.gdb_get_register(which))

    def write_register(self, data):
        reg = int(data.split(b'=')[0], 16)
        val = data.split(b'=')[1].split(b'#')[0]
        self.target_facade.set_register(reg, val)
        return self.create_rsp_packet(b"OK")

    def get_registers(self):
        return self.create_rsp_packet(self.target_facade.get_register_context())

    def set_registers(self, data):
        self.target_facade.set_register_context(data)
        return self.create_rsp_packet(b"OK")

    def handle_query(self, msg):
        query = msg.split(b':')
        LOG.debug('GDB received query: %s', query)

        if query is None:
            LOG.error('GDB received query packet malformed')
            return None

        if query[0] == b'Supported':
            # Save features sent by gdb.
            self.gdb_features = query[1].split(b';')

            # Build our list of features.
            features = [b'qXfer:features:read+', b'QStartNoAckMode+', b'qXfer:threads:read+', b'QNonStop+']
            features.append(b'PacketSize=' + (hex(self.packet_size).encode())[2:])
            if self.target_facade.get_memory_map_xml() is not None:
                features.append(b'qXfer:memory-map:read+')
            resp = b';'.join(features)
            return self.create_rsp_packet(resp)

        elif query[0] == b'Xfer':
            # qXfer:<object>:read:<annex>:<offset>,<length>
            if query[2] == b'read':
                data = query[4].split(b',')
                resp = self.handle_query_xml(query[1], query[3], int(data[0], 16), int(data[1].split(b'#')[0], 16))
                return self.create_rsp_packet(resp)
            else:
                LOG.debug("Unsupported qXfer request: %s:%s:%s:%s", query[1], query[2], query[3], query[4])
                # Must return an empty packet for an unrecognized qXfer.
                return self.create_rsp_packet(b"")

        elif query[0].startswith(b'C'):
            if not self.is_threading_enabled():
                return self.create_rsp_packet(b"QC1")
            else:
                self.validate_debug_context()
                return self.create_rsp_packet(("QC%x" % self.current_thread_id).encode())

        elif query[0].find(b'Attached') != -1:
            return self.create_rsp_packet(b"1")

        elif query[0].find(b'TStatus') != -1:
            return self.create_rsp_packet(b"")

        elif query[0].find(b'Tf') != -1:
            return self.create_rsp_packet(b"")

        elif b'Offsets' in query[0]:
            resp = b"Text=0;Data=0;Bss=0"
            return self.create_rsp_packet(resp)

        elif b'Symbol' in query[0]:
            if self.did_init_thread_providers:
                return self.create_rsp_packet(b"OK")
            return self.init_thread_providers()

        elif query[0].startswith(b'Rcmd,'):
            cmd = hex_decode(query[0][5:].split(b'#')[0])
            return self.handle_remote_command(cmd)

        else:
            return self.create_rsp_packet(b"")

    def init_thread_providers(self):
        if not self.session.options.get('rtos.enable'):
            LOG.debug("Skipping RTOS load because it was disabled.")
            return self.create_rsp_packet(b"OK")

        forced_rtos_name = self.session.options.get('rtos.name')
        if forced_rtos_name and (forced_rtos_name not in RTOS.keys()):
            LOG.error("%s was specified as the RTOS but no plugin with that name exists", forced_rtos_name)
            return self.create_rsp_packet(b"OK")

        symbol_provider = GDBSymbolProvider(self)

        LOG.info("Attempting to load RTOS plugins")
        for rtos_name, rtos_class in RTOS.items():
            if (forced_rtos_name is not None) and (rtos_name != forced_rtos_name):
                continue
            try:
                LOG.debug("Attempting to load %s", rtos_name)
                rtos = rtos_class(self.target)
                if rtos.init(symbol_provider):
                    LOG.info("%s loaded successfully", rtos_name)
                    self.thread_provider = rtos
                    break
                elif forced_rtos_name is not None:
                    LOG.error("%s was specified as the RTOS but failed to load", rtos_name)
            except exceptions.Error as e:
                LOG.error("Error during symbol lookup: %s", e, exc_info=self.session.log_tracebacks)

        self.did_init_thread_providers = True

        # Done with symbol processing.
        return self.create_rsp_packet(b"OK")

    def get_symbol(self, name: bytes) -> Optional[int]:
        assert self.packet_io

        # Send the symbol request.
        request = self.create_rsp_packet(b'qSymbol:' + hex_encode(name))
        self.packet_io.send(request)

        # Read a packet.
        packet = self.packet_io.receive()
        assert packet is not None

        # Parse symbol value reply packet.
        packet = packet[1:-3]
        if not packet.startswith(b'qSymbol:'):
            LOG.error("Got unexpected response from gdb when asking for symbol value")
            return None
        packet = packet[8:]
        sym_value, sym_name = packet.split(b':')

        sym_name = hex_decode(sym_name)
        if sym_name != name:
            LOG.error("Symbol value reply from gdb has unexpected symbol name (expected '%s', received '%s')",
                    name, sym_name)
            return None
        if sym_value:
            sym_value = hex8_to_u32le(sym_value)
        else:
            return None
        return sym_value

    def handle_remote_command(self, cmd):
        """@brief Pass remote commands to the commander command processor."""
        # Convert the command line to a string.
        cmd = to_str_safe(cmd)
        LOG.debug('Remote command: %s', cmd)

        # Create a new stream to collect the command output.
        stream = io.StringIO()
        self._command_context.output_stream = stream

        # TODO run this in a separate thread so we can cancel the command with ^C from gdb?
        try:
            # Run command and collect output.
            self._command_context.process_command_line(cmd)
        except exceptions.CommandError as err:
            stream.write("Error: %s\n" % err)
        except ToolExitException:
            stream.write("Error: cannot exit gdbserver\n")
        except exceptions.TransferError as err:
            stream.write("Transfer failed: %s\n" % err)
            LOG.error("Transfer failure while executing remote command '%s': %s", cmd, err,
                    exc_info=self.session.log_tracebacks)
        except Exception as err:
            stream.write("Unexpected error: %s\n" % err)
            LOG.error("Error while executing remote command '%s': %s", cmd, err,
                    exc_info=self.session.log_tracebacks)

        # Convert back to bytes, hex encode, then return the response packet.
        output = stream.getvalue()
        if not output:
            output = "OK\n"
        response = hex_encode(to_bytes_safe(output))

        # Disconnect the stream.
        self._command_context.output_stream = None

        return self.create_rsp_packet(response)

    def handle_general_set(self, msg):
        feature = msg.split(b'#')[0]
        LOG.debug("GDB general set: %s", feature)

        if feature == b'StartNoAckMode':
            # Disable acks after the reply and ack.
            self.packet_io.set_send_acks(False)
            return self.create_rsp_packet(b"OK")

        elif feature.startswith(b'NonStop'):
            enable = feature.split(b':')[1]
            self.non_stop = (enable == b'1')
            return self.create_rsp_packet(b"OK")

        else:
            return self.create_rsp_packet(b"")

    def handle_query_xml(self, query: bytes, annex: bytes, offset: int, size: int) -> bytes:
        LOG.debug('GDB query %s: annex: %s, offset: %s, size: %s', query, annex, offset, size)

        # For each query object, we check the annex and return E00 for invalid values. Only 'features'
        # has a non-empty annex.
        if query == b'memory-map':
            if annex != b'':
                return self.create_rsp_packet(b"E00")
            xml = self.target_facade.get_memory_map_xml()
        elif query == b'features':
            if annex == b'target.xml':
                xml = self.target_facade.get_target_xml()
            else:
                return self.create_rsp_packet(b"E00")
        elif query == b'threads':
            if annex != b'':
                return self.create_rsp_packet(b"E00")
            xml = self.get_threads_xml()
        else:
            # Unrecognised query object, so return empty packet.
            LOG.debug("Unsupported XML query (%s), annex (%s)", query, annex)
            return self.create_rsp_packet(b"")

        size_xml = len(xml)

        prefix = b'm'

        if offset > size_xml:
            LOG.error('GDB requested xml offset > size for %s!', query)
            return self.create_rsp_packet(b"E16") # EINVAL

        if size > (self.packet_size - 4):
            size = self.packet_size - 4

        nbBytesAvailable = size_xml - offset

        if size > nbBytesAvailable:
            prefix = b'l'
            size = nbBytesAvailable

        resp = prefix + escape(xml[offset:offset + size])

        return resp

    def create_rsp_packet(self, data):
        resp = b'$' + data + b'#' + checksum(data)
        return resp

    def syscall(self, op: str) -> Tuple[int, int]:
        LOG.debug("GDB server syscall: %s", op)
        request = self.create_rsp_packet(b'F' + op.encode())
        self.packet_io.send(request)

        while not self.packet_io.interrupt_event.is_set():
            # Read a packet.
            packet = self.packet_io.receive(False)
            if packet is None:
                sleep(0.1)
                continue

            # Check for file I/O response.
            if packet[0:1] == b'$' and packet[1:2] == b'F':
                LOG.debug("Syscall: got syscall response " + packet)
                args = packet[2:packet.index(b'#')].split(b',')
                result = int(args[0], base=16)
                errno = int(args[1], base=16) if len(args) > 1 else 0
                ctrl_c = args[2] if len(args) > 2 else b''
                if ctrl_c == b'C':
                    self.packet_io.interrupt_event.set()
                    self.packet_io.drop_reply = True
                return result, errno

            # decode and prepare resp
            resp, detach = self.handle_message(packet)

            if resp is not None:
                # send resp
                self.packet_io.send(resp)

            if detach:
                self.detach_event.set()
                LOG.warning("GDB server received detach request while waiting for file I/O completion")
                break

        return -1, 0

    def get_t_response(self, forceSignal=None):
        self.validate_debug_context()
        response = self.target_facade.get_t_response(forceSignal)

        # Append thread
        if not self.is_threading_enabled():
            response += b"thread:1;"
        else:
            if self.current_thread_id in (-1, 0, 1):
                response += ("thread:%x;" % self.thread_provider.current_thread.unique_id).encode()
            else:
                response += ("thread:%x;" % self.current_thread_id).encode()

        # Optionally append core
        if self.report_core:
            response += ("core:%x;" % self.core).encode()
        LOG.debug("Tresponse=%s", response)
        return response

    def get_threads_xml(self):
        root = Element('threads')

        if not self.is_threading_enabled():
            t = SubElement(root, 'thread', id="1")
            if self.report_core:
                t.set("core", str(self.core))
            if self.is_target_in_reset():
                t.text = "Reset"
            else:
                t.text = self.exception_name()
        else:
            threads = self.thread_provider.get_threads()
            for thread in threads:
                hexId = "%x" % thread.unique_id
                t = SubElement(root, 'thread', id=hexId, name=thread.name)
                if self.report_core:
                    t.set("core", str(self.core))
                t.text = thread.description

        return b'<?xml version="1.0"?><!DOCTYPE feature SYSTEM "threads.dtd">' + tostring(root)

    def is_threading_enabled(self):
        return (self.thread_provider is not None) and self.thread_provider.is_enabled \
            and (self.thread_provider.current_thread is not None)

    def is_target_in_reset(self):
        return self.target.get_state() == Target.State.RESET

    def exception_name(self):
        try:
            ipsr = self.target_context.read_core_register('ipsr')
            return self.target_context.core.exception_number_to_name(ipsr)
        except exceptions.Error:
            return None

    def event_handler(self, notification):
        if notification.event == Target.Event.POST_RESET:
            # Invalidate threads list if flash is reprogrammed.
            LOG.debug("Received POST_RESET event")
            self.first_run_after_reset_or_flash = True
            if self.thread_provider is not None:
                self.thread_provider.read_from_target = False

    def _option_did_change(self, notification):
        """@brief Handle an option changing at runtime.

        For option notifications, the event is the name of the option and the `data` attribute is an
        OptionChangeInfo object with `new_value` and `old_value` attributes.
        """
        if notification.event == 'vector_catch':
            self.target.set_vector_catch(convert_vector_catch(notification.data.new_value))
        elif notification.event == 'step_into_interrupt':
            self.step_into_interrupt = notification.data.new_value
        elif notification.event == 'persist':
            self.persist = notification.data.new_value
        elif notification.event == 'enable_semihosting':
            self.enable_semihosting = notification.data.new_value
            LOG.info("Semihosting %s", ('enabled' if self.enable_semihosting else 'disabled'))
        elif notification.event == 'report_core_number':
            self.report_core = notification.data.new_value




# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
import binascii
from typing import (Any, Iterator, List, Sequence, Tuple, cast)

from .mask import align_up

def byte_list_to_nbit_le_list(data: Sequence[int], bitwidth: int, pad: int = 0x00) -> List[int]:
    """@brief Convert a list of bytes to a list of n-bit integers (little endian)

    If the length of the data list is not a multiple of `bitwidth` // 8, then the pad value is used
    for the additional required bytes.

    @param data List of bytes.
    @param bitwidth Width in bits of the resulting values.
    @param pad Optional value used to pad input data if not aligned to the bitwidth.
    @result List of integer values that are `bitwidth` bits wide.
    """
    bytewidth = bitwidth // 8
    datalen = len(data) // bytewidth * bytewidth
    res = [sum((data[offset + i] << (i * 8)) for i in range(bytewidth))
            for offset in range(0, datalen, bytewidth)
            ]
    remainder = len(data) % bytewidth
    if remainder != 0:
        pad_count = bytewidth - remainder
        padded_data = list(data[-remainder:]) + [pad] * pad_count
        res.append(sum((padded_data[i] << (i * 8)) for i in range(bytewidth)))
    return res

def nbit_le_list_to_byte_list(data: Sequence[int], bitwidth: int) -> List[int]:
    """@brief Convert a list of n-bit values into a byte list.

    @param data List of n-bit values.
    @param bitwidth Width in bits of the input vales.
    @result List of integer bytes.
    """
    return [(x >> shift) & 0xff for x in data for shift in range(0, bitwidth, 8)]

def byte_list_to_u32le_list(data: Sequence[int], pad: int = 0x00) -> List[int]:
    """@brief Convert a list of bytes to a list of 32-bit integers (little endian)

    If the length of the data list is not a multiple of 4, then the pad value is used
    for the additional required bytes.
    """
    res = []
    for i in range(len(data) // 4):
        res.append(data[i * 4 + 0] |
                   data[i * 4 + 1] << 8 |
                   data[i * 4 + 2] << 16 |
                   data[i * 4 + 3] << 24)
    remainder = (len(data) % 4)
    if remainder != 0:
        pad_count = 4 - remainder
        res += byte_list_to_u32le_list(list(data[-remainder:]) + [pad] * pad_count)
    return res

def u32le_list_to_byte_list(data: Sequence[int]) -> List[int]:
    """@brief Convert a word array into a byte array"""
    res = []
    for x in data:
        res.append((x >> 0) & 0xff)
        res.append((x >> 8) & 0xff)
        res.append((x >> 16) & 0xff)
        res.append((x >> 24) & 0xff)
    return res

def u16le_list_to_byte_list(data: Sequence[int]) -> List[int]:
    """@brief Convert a halfword array into a byte array"""
    byte_data = []
    for h in data:
        byte_data.extend([h & 0xff, (h >> 8) & 0xff])
    return byte_data

def byte_list_to_u16le_list(byte_data: Sequence[int]) -> List[int]:
    """@brief Convert a byte array into a halfword array"""
    data = []
    for i in range(0, len(byte_data), 2):
        data.append(byte_data[i] | (byte_data[i + 1] << 8))
    return data

def u32_to_float32(data: int) -> float:
    """@brief Convert a 32-bit int to an IEEE754 float"""
    d = struct.pack(">I", data & 0xffff_ffff)
    return struct.unpack(">f", d)[0]

def float32_to_u32(data: float) -> int:
    """@brief Convert an IEEE754 float to a 32-bit int"""
    d = struct.pack(">f", data)
    return struct.unpack(">I", d)[0]

def u64_to_float64(data: int) -> float:
    """@brief Convert a 64-bit int to an IEEE754 float"""
    d = struct.pack(">Q", data & 0xffff_ffff_ffff_ffff)
    return struct.unpack(">d", d)[0]

def float64_to_u64(data: float) -> int:
    """@brief Convert an IEEE754 float to a 64-bit int"""
    d = struct.pack(">d", data)
    return struct.unpack(">Q", d)[0]

def uint_to_hex_le(value: int, width: int) -> str:
    """@brief Create an n-digit hexadecimal string from an integer value.
    @param value Integer value to format.
    @param width The width in bits.
    @return A string with the number of hex bytes required to fit `width` bits, rounded up to the
        next whole byte. The bytes represent `value` in little-endian order. That is, the first hex
        byte contains the LSB of `value`, while the last hex byte the MSB.
    """
    return ''.join("%02x" % ((value >> b) & 0xff) for b in range(0, align_up(width, 8), 8))

def hex_le_to_uint(value: str, width: int) -> int:
    """@brief Create an an integer value from an n-digit hexadecimal string.
    @param value String consisting of pairs of hex digits with no intervening whitespace. Must have at least
        enough hex bytes to meet the desired width. The first hex byte is the LSB.
    @param width The width in bits. The width can be shorter then the input `value` width, in which case
        more significant bytes will be truncated.
    @return An integer converted from `value`.
    """
    return sum((int(value[i:i+2], base=16) << (i * 4)) for i in range(0, align_up(width, 8) // 4, 2))

def u32_to_hex8le(val: int) -> str:
    """@brief Create 8-digit hexadecimal string from 32-bit register value"""
    return uint_to_hex_le(val, 32)

def u64_to_hex16le(val: int) -> str:
    """@brief Create 16-digit hexadecimal string from 64-bit register value"""
    return uint_to_hex_le(val, 64)

def hex8_to_u32be(data: str) -> int:
    """@brief Build 32-bit register value from big-endian 8-digit hexadecimal string
    @note Endianness in this function name is backwards.
    """
    return hex_le_to_uint(data, 32)

def hex16_to_u64be(data: str) -> int:
    """@brief Build 64-bit register value from big-endian 16-digit hexadecimal string
    @note Endianness in this function name is backwards.
    """
    return hex_le_to_uint(data, 64)

def hex8_to_u32le(data: str) -> int:
    """@brief Build 32-bit register value from little-endian 8-digit hexadecimal string
    @note Endianness in this function name is backwards.
    """
    return int(data[0:8], 16)

def hex16_to_u64le(data: str) -> int:
    """@brief Build 64-bit register value from little-endian 16-digit hexadecimal string
    @note Endianness in this function name is backwards.
    """
    return int(data[0:16], 16)

def byte_to_hex2(val: int) -> str:
    """@brief Create 2-digit hexadecimal string from 8-bit value"""
    return "%02x" % int(val)

def hex_to_byte_list(data: str) -> List[int]:
    """@brief Convert string of hex bytes to list of integers"""
    return list(binascii.unhexlify(data))

def hex_decode(cmd: str) -> bytes:
    """@brief Return the binary data represented by the hexadecimal string."""
    return binascii.unhexlify(cmd)

def hex_encode(string: bytes) -> bytes:
    """@brief Return the hexadecimal representation of the binary data."""
    return binascii.hexlify(string)

def pairwise(iterable: Iterator[Any]) -> Iterator[Tuple[Any, Any]]:
    """s -> (s0,s1), (s2,s3), (s3, s4), ..."""
    r = []
    for x in iterable:
        r.append(x)
        if len(r) == 2:
            yield cast(Tuple[Any, Any], tuple(r))
            r = []



# pyOCD debugger
# Copyright (c) 2020 NXP
# Copyright (c) 2006-2013,2018 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ..family.target_kinetis import Kinetis
from ..family.flash_kinetis import Flash_Kinetis
from ...core.memory_map import (FlashRegion, RamRegion, RomRegion, MemoryMap)
from ...coresight import ap
from ...debug.svd.loader import SVDFile

LOG = logging.getLogger(__name__)

SIM_SDID = 0x40075024
SIM_SDID_KEYATTR_MASK = 0x70
SIM_SDID_KEYATTR_SHIFT = 4

KEYATTR_DUAL_CORE = 1

RCM_MR = 0x4007f010
RCM_MR_BOOTROM_MASK = 0x6

SCG_CSR = 0x4007B010

SCG_RCCR = 0x4007B014
SCS_MASK = 0x0F000000
SCS_SHIFT = 24
DIVCORE_MASK = 0x000F0000
DIVCORE_SHIFT = 16
DIVSLOW_MASK = 0x0000000F
DIVSLOW_SHIFT = 0

SCG_FIRCCSR = 0x4007B300
FIRCEN_MASK = 1

SCG_FIRCCFG = 0x4007B308

RECOVER_TIMEOUT = 1.0 # 1 second

FLASH_ALGO = {
    'load_address' : 0x20000000,
    'instructions' : [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x4832b510, 0x60414930, 0x60814931, 0x22806801, 0x22204391, 0x60014311, 0x4448482e, 0xf860f000,
    0xd0002800, 0xbd102001, 0x47702000, 0xb5104829, 0x44484929, 0xf922f000, 0xd1042800, 0x21004825,
    0xf0004448, 0x4a25f9c3, 0x230168d1, 0x4319029b, 0xbd1060d1, 0x4c1fb570, 0x444c4605, 0x4b1e4601,
    0x68e24620, 0xf89ef000, 0xd1052800, 0x46292300, 0x68e24620, 0xf916f000, 0x68ca4918, 0x029b2301,
    0x60ca431a, 0xb570bd70, 0x460b460c, 0x46014606, 0xb0844810, 0x44484615, 0xf8bef000, 0xd10a2800,
    0x90029001, 0x480b2101, 0x462b9100, 0x46314622, 0xf0004448, 0x4909f957, 0x230168ca, 0x431a029b,
    0xb00460ca, 0x0000bd70, 0xd928c520, 0x40076000, 0x0000ffff, 0x00000004, 0x6b65666b, 0xf0003000,
    0xd00a2800, 0x68c9492b, 0x0e094a2b, 0x447a0049, 0x03095a51, 0x2064d103, 0x20044770, 0xb4104770,
    0x60032300, 0x21026041, 0x02896081, 0x492360c1, 0x158a7a0c, 0x610240a2, 0x61837ac9, 0xbc106141,
    0x47704618, 0xd0022800, 0x20006181, 0x20044770, 0x28004770, 0x2004d101, 0xb4104770, 0x42191e5b,
    0x421ad101, 0xbc10d002, 0x47702065, 0x428b6803, 0x6840d804, 0x18181889, 0xd2024288, 0x2066bc10,
    0xbc104770, 0x47702000, 0x4288490d, 0x206bd001, 0x20004770, 0x28004770, 0x290fd008, 0x2a04d802,
    0xe005d104, 0xd8012913, 0xd0012a08, 0x47702004, 0x47702000, 0x40075040, 0x00000512, 0x40020020,
    0x6b65666b, 0x4605b5f8, 0x460c4616, 0xf7ff4618, 0x2800ffdb, 0x2308d12b, 0x46214632, 0xf7ff4628,
    0x0007ffb8, 0x19a6d123, 0x1e7668e9, 0x91004630, 0xf922f000, 0xd0032900, 0x1c409e00, 0x1e764346,
    0xd81342b4, 0x4478480a, 0x60046800, 0x20094909, 0xf00071c8, 0x4607f8f9, 0x280069a8, 0x4780d000,
    0xd1032f00, 0x190468e8, 0xd9eb42b4, 0xbdf84638, 0x00000416, 0x40020000, 0xd1012a00, 0x47702004,
    0x461cb5ff, 0x4615b081, 0x2304460e, 0x98014622, 0xff7ff7ff, 0xd11a0007, 0xd0172c00, 0x4478480d,
    0x600e6801, 0x6800cd02, 0x490b6041, 0x71c82006, 0xf8caf000, 0x98014607, 0x28006980, 0x4780d000,
    0xd1032f00, 0x1d361f24, 0xd1e72c00, 0xb0054638, 0x0000bdf0, 0x000003be, 0x40020000, 0x4604b510,
    0xf7ff4608, 0x2800ff71, 0x2c00d106, 0x4904d005, 0x71c82044, 0xf8a8f000, 0x2004bd10, 0x0000bd10,
    0x40020000, 0xb081b5ff, 0x460e4614, 0x23084605, 0xff3ff7ff, 0xd12a2800, 0x686868a9, 0xf8acf000,
    0x42719000, 0x40014240, 0x42b7424f, 0x9800d101, 0x2c00183f, 0x1bbdd01a, 0xd90042a5, 0x490d4625,
    0x447908e8, 0x600e6809, 0x2201490b, 0x0a0271ca, 0x728872ca, 0x72489804, 0xf876f000, 0xd1062800,
    0x1b649800, 0x183f1976, 0xd1e42c00, 0xb0052000, 0x0000bdf0, 0x0000031a, 0x40020000, 0xd00c2800,
    0xd00a2a00, 0xd21a2908, 0x447b000b, 0x18db791b, 0x0705449f, 0x0d0b0907, 0x2004110f, 0x68c04770,
    0x6840e00a, 0x6880e008, 0x6800e006, 0x2001e004, 0x6900e002, 0x6940e000, 0x20006010, 0x206a4770,
    0x00004770, 0xd1012b00, 0x47702004, 0x461cb5f8, 0x460e4615, 0x9f082304, 0xfedbf7ff, 0xd1192800,
    0xd0172d00, 0x447a4a0f, 0x60066810, 0x2102480e, 0x990671c1, 0x681172c1, 0x60886820, 0xf824f000,
    0xd0082800, 0x29009907, 0x600ed000, 0xd0012f00, 0x60392100, 0x1d24bdf8, 0x1d361f2d, 0xd1e12d00,
    0x0000bdf8, 0x00000276, 0x40020000, 0xd1012800, 0x47702004, 0x4803b510, 0x71c22240, 0xf0007181,
    0xbd10f803, 0x40020000, 0x2170480a, 0x21807001, 0x78017001, 0xd5fc0609, 0x06817800, 0x2067d501,
    0x06c14770, 0x2068d501, 0x07c04770, 0x2069d0fc, 0x00004770, 0x40020000, 0x09032200, 0xd373428b,
    0x428b0a03, 0x0b03d358, 0xd33c428b, 0x428b0c03, 0xe012d321, 0x430b4603, 0x2200d47f, 0x428b0843,
    0x0903d374, 0xd35f428b, 0x428b0a03, 0x0b03d344, 0xd328428b, 0x428b0c03, 0x22ffd30d, 0xba120209,
    0x428b0c03, 0x1212d302, 0xd0650209, 0x428b0b03, 0xe000d319, 0x0bc30a09, 0xd301428b, 0x1ac003cb,
    0x0b834152, 0xd301428b, 0x1ac0038b, 0x0b434152, 0xd301428b, 0x1ac0034b, 0x0b034152, 0xd301428b,
    0x1ac0030b, 0x0ac34152, 0xd301428b, 0x1ac002cb, 0x0a834152, 0xd301428b, 0x1ac0028b, 0x0a434152,
    0xd301428b, 0x1ac0024b, 0x0a034152, 0xd301428b, 0x1ac0020b, 0xd2cd4152, 0x428b09c3, 0x01cbd301,
    0x41521ac0, 0x428b0983, 0x018bd301, 0x41521ac0, 0x428b0943, 0x014bd301, 0x41521ac0, 0x428b0903,
    0x010bd301, 0x41521ac0, 0x428b08c3, 0x00cbd301, 0x41521ac0, 0x428b0883, 0x008bd301, 0x41521ac0,
    0x428b0843, 0x004bd301, 0x41521ac0, 0xd2001a41, 0x41524601, 0x47704610, 0x0fcae05d, 0x4249d000,
    0xd3001003, 0x40534240, 0x469c2200, 0x428b0903, 0x0a03d32d, 0xd312428b, 0x018922fc, 0x0a03ba12,
    0xd30c428b, 0x11920189, 0xd308428b, 0x11920189, 0xd304428b, 0xd03a0189, 0xe0001192, 0x09c30989,
    0xd301428b, 0x1ac001cb, 0x09834152, 0xd301428b, 0x1ac0018b, 0x09434152, 0xd301428b, 0x1ac0014b,
    0x09034152, 0xd301428b, 0x1ac0010b, 0x08c34152, 0xd301428b, 0x1ac000cb, 0x08834152, 0xd301428b,
    0x1ac0008b, 0xd2d94152, 0x428b0843, 0x004bd301, 0x41521ac0, 0xd2001a41, 0x46634601, 0x105b4152,
    0xd3014610, 0x2b004240, 0x4249d500, 0x46634770, 0xd300105b, 0xb5014240, 0x46c02000, 0xbd0246c0,
    0x40020004, 0x00000000, 0x00000000, 0x00100000, 0x00200000, 0x00400000, 0x00800000, 0x00000000,
    0x00800000, 0x00000000,
    ],

    'pc_init' : 0x20000021,
    'pc_unInit': 0x20000049,
    'pc_program_page': 0x200000A7,
    'pc_erase_sector': 0x20000075,
    'pc_eraseAll' : 0x2000004D,

    'static_base' : 0x20000000 + 0x00000020 + 0x00000624,
    'begin_stack' : 0x20000000 + 0x00000800,
    'begin_data' : 0x20000000 + 0x00000A00,
    'page_size' : 0x00000200,

    # All keys above are auto-generated. The following are added or modified.
    'analyzer_supported' : True,                 # [modified] default is False
    'analyzer_address' : 0x1fffa000,             # [modified] default is zero. Use 8K block before flash algo. Can be any unused SRAM.
    'page_buffers' : [0x20000a00, 0x20001200],   # [added] Use areas above algo. Note 'begin_data' is unused if double buffering. Can be any unused SRAM.
    'min_program_length' : 4                     # [added] See FSL_FEATURE_FLASH_PFLASH_BLOCK_WRITE_UNIT_SIZE in KSDK features header file
}

class Flash_kl28z(Flash_Kinetis):
    def __init__(self, target):
        super(Flash_kl28z, self).__init__(target, FLASH_ALGO)
        self._saved_firccsr = 0
        self._saved_rccr = 0

    def prepare_target(self):
        """This function sets up target clocks to ensure that flash is clocked at the maximum
        of 24MHz. Doing so gets the best flash programming performance. The FIRC clock source
        is used so that there is no dependency on an external crystal frequency.
        """
        # Enable FIRC.
        value = self.target.read32(SCG_FIRCCSR)
        self._saved_firccsr = value
        value |= FIRCEN_MASK
        self.target.write32(SCG_FIRCCSR, value)

        # Switch system to FIRC, core=48MHz (/1), slow=24MHz (/2).
        # Flash and the bus are clocked from the slow clock, and its max is 24MHz,
        # so there is no benefit from raising the core clock further.
        self._saved_rccr = self.target.read32(SCG_RCCR)
        self.target.write32(SCG_RCCR, (0x3 << SCS_SHIFT) | (1 << DIVSLOW_SHIFT))

        csr = self.target.read32(SCG_CSR)
        LOG.debug("SCG_CSR = 0x%08x", csr)

    def restore_target(self):
        """Restore clock registers to original values."""
        self.target.write32(SCG_FIRCCSR, self._saved_firccsr)
        self.target.write32(SCG_RCCR, self._saved_rccr)


class KL28x(Kinetis):

    SINGLE_MAP = MemoryMap(
        FlashRegion(name='flash', start=0, length=0x80000, blocksize=0x800, is_boot_memory=True,
            flash_class=Flash_kl28z,
            algo=FLASH_ALGO),
        RamRegion(name='ram', start=0x1fff8000, length=0x20000),
        RamRegion(name='usb ram', start=0x40100000, length=0x800)
        )

    DUAL_MAP = MemoryMap(
        FlashRegion(name='flash', start=0, length=0x80000, blocksize=0x800, is_boot_memory=True,
            flash_class=Flash_kl28z,
            algo=FLASH_ALGO),
        RomRegion(name='core1 imem alias', start=0x1d200000, length=0x40000),
        RamRegion(name='core0 ram', start=0x1fffa000, length=0x18000),
        RomRegion(name='core1 imem', start=0x2d200000, length=0x40000),
        RamRegion(name='core1 dmem', start=0x2d300000, length=0x8000),
        RamRegion(name='usb ram', start=0x40100000, length=0x800)
        )

    def __init__(self, session):
        super(KL28x, self).__init__(session, self.SINGLE_MAP)
        self.is_dual_core = False

        self._svd_location = SVDFile.from_builtin("MKL28T7_CORE0.svd")

    def create_init_sequence(self):
        seq = super(KL28x, self).create_init_sequence()

        seq.wrap_task('discovery',
            lambda seq: seq
                # The KL28 will lock up if an invalid AP is accessed, so replace the AP scan with a
                # fixed list of known APs.
                .replace_task('find_aps', self.create_kl28_aps)
                # Before creating cores, determine which memory map should be used.
                .insert_before('create_cores',
                    ('detect_dual_core', self.detect_dual_core)
                    )
            )

        return seq

    def create_kl28_aps(self):
        """@brief Set the fixed list of valid AP numbers for KL28."""
        self.dp.valid_aps = [0, 1, 2]

    def detect_dual_core(self):
        if not isinstance(self.aps[0], ap.MEM_AP):
            return

        # Check if this is the dual core part.
        sdid = self.aps[0].read_memory(SIM_SDID)
        keyattr = (sdid & SIM_SDID_KEYATTR_MASK) >> SIM_SDID_KEYATTR_SHIFT
        LOG.debug("KEYATTR=0x%x SDID=0x%08x", keyattr, sdid)
        self.is_dual_core = (keyattr == KEYATTR_DUAL_CORE)
        if self.is_dual_core:
            LOG.info("KL28 is dual core")
            self.memory_map = self.DUAL_MAP

    def post_connect_hook(self):
        if not isinstance(self.aps[0], ap.MEM_AP):
            return
        # Disable ROM vector table remapping.
        self.aps[0].write32(RCM_MR, RCM_MR_BOOTROM_MASK)







# Copyright (c) 2018-2019, Arm Limited and affiliates.
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from abc import ABCMeta, abstractmethod
from io import open
from os import listdir
from os.path import join, exists, isdir
import logging

LOG = logging.getLogger(__name__)

class StlinkDetectBase(object, metaclass=ABCMeta):
    """ Base class for stlink detection, defines public interface for
    mbed-enabled stlink devices detection for various hosts
    """

    MBED_HTM_NAME = "mbed.htm"

    def __init__(self):
        """ ctor
        """
        pass

    @abstractmethod
    def find_candidates(self):
        """Find all candidate devices connected to this computer

        Note: Should not open any files

        @return A dict with the keys 'mount_point', 'serial_port' and 'target_id_usb_id'
        """
        raise NotImplementedError

    def list_mbeds(self):
        """ List details of connected devices
        @return Returns list of structures with detailed info about each mbed
        @details Function returns list of dictionaries with mbed attributes
          'mount_point', TargetID name etc.
        Function returns mbed list with platform names if possible
        """
        candidates = list(self.find_candidates())
        result = []
        for device in candidates:
            if not device.get("mount_point", None):
                continue
            device["target_id"] = device["target_id_usb_id"]
            self._update_device_from_fs(device)
            result.append(device)

        return result

    def _update_device_from_fs(self, device):
        """ Updates the device information based on files from its 'mount_point'
            @param device Dictionary containing device information
        """
        try:
            directory_entries = listdir(device["mount_point"])

            # Always try to update using daplink compatible boards processself.
            # This is done for backwards compatibility.
            lowercase_directory_entries = [e.lower() for e in directory_entries]
            if self.MBED_HTM_NAME.lower() in lowercase_directory_entries:
                self._update_device_from_htm(device)

        except (OSError, IOError) as e:
            LOG.warning(
                'Marking device with mount point "%s" as unmounted due to the '
                "following error: %s",
                device["mount_point"],
                e,
            )
            device["mount_point"] = None

    def _update_device_from_htm(self, device):
        """Set the 'target_id', 'target_id_mbed_htm', 'platform_name' and
        'daplink_*' attributes by reading from mbed.htm on the device
        """
        htm_target_id, daplink_info = self._read_htm_ids(device["mount_point"])
        if htm_target_id:
            device["target_id"] = htm_target_id
        else:
            LOG.debug(
                "Could not read htm on from usb id %s. Falling back to usb id",
                device["target_id_usb_id"],
            )
            device["target_id"] = device["target_id_usb_id"]
        device["target_id_mbed_htm"] = htm_target_id

    # Private functions supporting API
    def _read_htm_ids(self, mount_point):
        """Function scans mbed.htm to get information about TargetID.
        @param mount_point mbed mount point (disk / drive letter)
        @return Function returns targetID, in case of failure returns None.
        @details Note: This function should be improved to scan variety of boards'
          mbed.htm files
        """
        result = {}
        target_id = None
        for line in self._htm_lines(mount_point):
            target_id = target_id or self._target_id_from_htm(line)
        return target_id, result

    def _htm_lines(self, mount_point):
        if mount_point:
            mbed_htm_path = join(mount_point, self.MBED_HTM_NAME)
            with open(mbed_htm_path, "r") as f:
                return f.readlines()

    def _target_id_from_htm(self, line):
        """Extract Target id from htm line.
        @return Target id or None
        """
        # Detecting modern mbed.htm file format
        m = re.search("\\?code=([a-fA-F0-9]+)", line)
        if m:
            result = m.groups()[0]
            return result
        # Last resort, we can try to see if old mbed.htm format is there
        m = re.search("\\?auth=([a-fA-F0-9]+)", line)
        if m:
            result = m.groups()[0]
            return result

        return None

    def mount_point_ready(self, path):
        """Check if a mount point is ready for file operations
        """
        return exists(path) and isdir(path)

    @staticmethod
    def _run_cli_process(cmd, shell=True):
        """Runs command as a process and return stdout, stderr and ret code
        @param cmd Command to execute
        @return Tuple of (stdout, stderr, returncode)
        """
        from subprocess import Popen, PIPE

        p = Popen(cmd, shell=shell, stdout=PIPE, stderr=PIPE)
        _stdout, _stderr = p.communicate()
        return _stdout, _stderr, p.returncode




# pyOCD debugger
# Copyright (c) 2020 Cypress Semiconductor Corporation
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .component import CoreSightCoreComponent
from ..core import exceptions
from ..core.target import Target
from ..core.core_registers import CoreRegistersIndex

LOG = logging.getLogger(__name__)

DEAD_VALUE = 0


class GenericMemAPTarget(Target, CoreSightCoreComponent):
    """@brief This target represents ARM debug Access Port without a CPU

    It may be used to access the address space of the target via Access Ports
    without real ARM CPU core behind it. For instance Cypress PSoC64 devices have
    three APs implemented in the hardware:
    * AP #0 -> CPU-less AHB AP
    * AP #1 -> Cortex-M0+ AP
    * AP #2 -> Cortex-M4F AP
    Depending on the protection state, AP #1 and AP #2 can be permanently disabled.
    This class allows to communicate with Secure FW running on the target via AP #0.

    Most of the methods in this class (except memory access methods) are empty/dummy.
    """

    def __init__(self, session, ap, memory_map=None, core_num=0, cmpid=None, address=None):
        Target.__init__(self, session, memory_map)
        CoreSightCoreComponent.__init__(self, ap, cmpid, address)
        self.core_number = core_num
        self.core_type = DEAD_VALUE
        self._core_registers = CoreRegistersIndex()
        self._target_context = None

    def add_child(self, cmp):
        pass

    @property
    def core_registers(self):
        return self._core_registers

    @property
    def supported_security_states(self):
        return Target.SecurityState.NONSECURE,

    def init(self):
        pass

    def disconnect(self, resume=True):
        pass

    def write_memory(self, addr, value, transfer_size=32):
        self.ap.write_memory(addr, value, transfer_size)

    def read_memory(self, addr, transfer_size=32, now=True):
        return self.ap.read_memory(addr, transfer_size, now)

    def read_memory_block8(self, addr, size):
        return self.ap.read_memory_block8(addr, size)

    def write_memory_block8(self, addr, data):
        self.ap.write_memory_block8(addr, data)

    def write_memory_block32(self, addr, data):
        self.ap.write_memory_block32(addr, data)

    def read_memory_block32(self, addr, size):
        return self.ap.read_memory_block32(addr, size)

    def halt(self):
        pass

    def step(self, disable_interrupts=True, start=0, end=0, hook_cb=None):
        pass

    def reset(self, reset_type=None):
        pass

    def reset_and_halt(self, reset_type=None):
        self.reset(reset_type)

    def get_state(self):
        return Target.State.HALTED

    def get_security_state(self):
        return Target.SecurityState.NONSECURE

    def is_running(self):
        return self.get_state() == Target.State.RUNNING

    def is_halted(self):
        return self.get_state() == Target.State.HALTED

    def resume(self):
        pass

    def find_breakpoint(self, addr):
        return None

    def read_core_register(self, reg):
        raise exceptions.CoreRegisterAccessError("GenericMemAPTarget does not support core register access")

    def read_core_register_raw(self, reg):
        raise exceptions.CoreRegisterAccessError("GenericMemAPTarget does not support core register access")

    def read_core_registers_raw(self, reg_list):
        raise exceptions.CoreRegisterAccessError("GenericMemAPTarget does not support core register access")

    def write_core_register(self, reg, data):
        raise exceptions.CoreRegisterAccessError("GenericMemAPTarget does not support core register access")

    def write_core_register_raw(self, reg, data):
        raise exceptions.CoreRegisterAccessError("GenericMemAPTarget does not support core register access")

    def write_core_registers_raw(self, reg_list, data_list):
        raise exceptions.CoreRegisterAccessError("GenericMemAPTarget does not support core register access")

    def set_breakpoint(self, addr, type=Target.BreakpointType.AUTO):
        return False

    def remove_breakpoint(self, addr):
        pass

    def get_breakpoint_type(self, addr):
        return None

    def set_watchpoint(self, addr, size, type):
        return False

    def remove_watchpoint(self, addr, size, type):
        pass

    def set_vector_catch(self, enable_mask):
        pass

    def get_vector_catch(self):
        return 0

    def get_halt_reason(self):
        return Target.HaltReason.DEBUG

    def get_target_context(self, core=None):
        return self._target_context

    def set_target_context(self, context):
        self._target_context = context

    def create_init_sequence(self):
        pass

    def mass_erase(self):
        pass



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import List
import logging

from .base import SubcommandBase
from ..core.helpers import ConnectHelper
from ..flash.eraser import FlashEraser
from ..utility.cmdline import (
    convert_session_options,
    flatten_args,
)

LOG = logging.getLogger(__name__)

class EraseSubcommand(SubcommandBase):
    """@brief `pyocd erase` subcommand."""

    NAMES = ['erase']
    HELP = "Erase entire device flash or specified sectors."
    EPILOG = ("If no position arguments are listed, then no action will be taken unless the --chip or "
            "--mass-erase options are provided. Otherwise, the positional arguments should be the addresses of flash "
            "sectors or address ranges. The end address of a range is exclusive, meaning that it will not be "
            "erased. Thus, you should specify the address of the sector after the last one "
            "to be erased. If a '+' is used instead of '-' in a range, this indicates that the "
            "second value is a length rather than end address. "
            "Examples: 0x1000 (erase single sector starting at 0x1000) "
            "0x800-0x2000 (erase sectors starting at 0x800 up to but not including 0x2000) "
            "0+8192 (erase 8 kB starting at address 0)")
    DEFAULT_LOG_LEVEL = logging.WARNING

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""
        erase_parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        erase_options = erase_parser.add_argument_group("erase options")
        erase_options.add_argument("-c", "--chip", dest="erase_mode", action="store_const", const=FlashEraser.Mode.CHIP,
            help="Perform a chip erase.")
        erase_options.add_argument("-s", "--sector", dest="erase_mode", action="store_const", const=FlashEraser.Mode.SECTOR,
            help="Erase the sectors listed as positional arguments. This is the default.")
        erase_options.add_argument("--mass", dest="erase_mode", action="store_const", const=FlashEraser.Mode.MASS,
            help="Perform a mass erase. On some devices this is different than a chip erase.")

        erase_parser.add_argument("addresses", metavar="<sector-address>", action='append', nargs='*',
            help="List of sector addresses or ranges to erase.")

        return [cls.CommonOptions.COMMON, cls.CommonOptions.CONNECT, erase_parser]

    def invoke(self) -> int:
        """@brief Handle 'erase' subcommand."""
        self._increase_logging(["pyocd.flash.eraser"])

        # Display a nice, helpful error describing why nothing was done and how to correct it.
        if (self._args.erase_mode is None) or not self._args.addresses:
            LOG.error("No erase operation specified. Please specify one of '--chip', '--sector', "
                        "or '--mass' to indicate the desired erase mode. For sector erases, a list "
                        "of sector addresses to erase must be provided. "
                        "See 'pyocd erase --help' for more.")
            return 1

        session = ConnectHelper.session_with_chosen_probe(
                            project_dir=self._args.project_dir,
                            config_file=self._args.config,
                            user_script=self._args.script,
                            no_config=self._args.no_config,
                            pack=self._args.pack,
                            unique_id=self._args.unique_id,
                            target_override=self._args.target_override,
                            frequency=self._args.frequency,
                            blocking=(not self._args.no_wait),
                            connect_mode=self._args.connect_mode,
                            options=convert_session_options(self._args.options),
                            option_defaults=self._modified_option_defaults(),
                            )
        if session is None:
            LOG.error("No device available to erase")
            return 1
        with session:
            mode = self._args.erase_mode or FlashEraser.Mode.SECTOR
            eraser = FlashEraser(session, mode)

            addresses = flatten_args(self._args.addresses)
            eraser.erase(addresses)

        return 0





# pyOCD debugger
# Copyright (c) 2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core.target import Target

## Map from DFP reset sequence names to our reset types.
RESET_SEQUENCE_TO_TYPE_MAP = {
    'ResetHardware':    Target.ResetType.HW,
    'ResetSystem':      Target.ResetType.SW_SYSTEM,
    'ResetProcessor':   Target.ResetType.SW_CORE,
}

## Map from DFP reset sequence names to our reset types.
#
# ResetType.SW is expected to have been mapped to its actual value before using this map.
# ResetType.SW_EMULATED doesn't have a corresponding reset sequence, so it must be handled
# in another way.
RESET_TYPE_TO_SEQUENCE_MAP = {
    Target.ResetType.HW:        'ResetHardware',
    Target.ResetType.SW_SYSTEM: 'ResetSystem',
    Target.ResetType.SW_CORE:   'ResetProcessor',
}
